from __future__ import annotations

import os
import random
import shutil
import tempfile
from contextlib import contextmanager
from dataclasses import dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING, Any, Generator, Optional, TypeVar
from uuid import NAMESPACE_DNS, uuid5

from pydantic import BaseModel, Field, Secret

from unstructured_ingest.v2.interfaces import (
    AccessConfig,
    ConnectionConfig,
    Downloader,
    DownloaderConfig,
    DownloadResponse,
    Indexer,
    IndexerConfig,
    Uploader,
    UploaderConfig,
)
from unstructured_ingest.v2.logger import logger
from unstructured_ingest.v2.processes.connectors.fsspec.utils import sterilize_dict
from unstructured_ingest.v2.types.file_data import (
    FileData,
    FileDataSourceMetadata,
    SourceIdentifiers,
)

if TYPE_CHECKING:
    from fsspec import AbstractFileSystem

CONNECTOR_TYPE = "fsspec"


class FileConfig(BaseModel):
    remote_url: str = Field(description="Remote fsspec URL formatted as `protocol://dir/path`")
    protocol: str = Field(init=False)
    path_without_protocol: str = Field(init=False)
    supported_protocols: list[str] = Field(
        init=False,
        default_factory=lambda: [
            "s3",
            "s3a",
            "abfs",
            "az",
            "gs",
            "gcs",
            "box",
            "dropbox",
            "sftp",
        ],
    )

    def __init__(self, **data):
        protocol, path_without_protocol = data["remote_url"].split("://")
        data["protocol"] = protocol
        data["path_without_protocol"] = path_without_protocol
        super().__init__(**data)


class FsspecIndexerConfig(FileConfig, IndexerConfig):
    recursive: bool = False
    sample_n_files: Optional[int] = None


class FsspecAccessConfig(AccessConfig):
    pass


class FsspecConnectionConfig(ConnectionConfig):
    access_config: Secret[FsspecAccessConfig]
    connector_type: str = Field(default=CONNECTOR_TYPE, init=False)

    @contextmanager
    def get_client(self, protocol: str) -> Generator["AbstractFileSystem", None, None]:
        from fsspec import get_filesystem_class

        client = get_filesystem_class(protocol)(
            **self.get_access_config(),
        )
        yield client

    def wrap_error(self, e: Exception) -> Exception:
        return e


FsspecIndexerConfigT = TypeVar("FsspecIndexerConfigT", bound=FsspecIndexerConfig)
FsspecConnectionConfigT = TypeVar("FsspecConnectionConfigT", bound=FsspecConnectionConfig)


@dataclass
class FsspecIndexer(Indexer):
    connection_config: FsspecConnectionConfigT
    index_config: FsspecIndexerConfigT
    connector_type: str = Field(default=CONNECTOR_TYPE, init=False)

    def wrap_error(self, e: Exception) -> Exception:
        return self.connection_config.wrap_error(e=e)

    def precheck(self) -> None:
        from fsspec import get_filesystem_class

        try:
            fs = get_filesystem_class(self.index_config.protocol)(
                **self.connection_config.get_access_config(),
            )
            files = fs.ls(path=self.index_config.path_without_protocol, detail=True)
            valid_files = [x.get("name") for x in files if x.get("type") == "file"]
            if not valid_files:
                return
            file_to_sample = valid_files[0]
            logger.debug(f"attempting to make HEAD request for file: {file_to_sample}")
            with self.connection_config.get_client(protocol=self.index_config.protocol) as client:
                client.head(path=file_to_sample)
        except Exception as e:
            logger.error(f"failed to validate connection: {e}", exc_info=True)
            raise self.wrap_error(e=e)

    def get_file_info(self) -> list[dict[str, Any]]:
        if not self.index_config.recursive:
            # fs.ls does not walk directories
            # directories that are listed in cloud storage can cause problems
            # because they are seen as 0 byte files
            with self.connection_config.get_client(protocol=self.index_config.protocol) as client:
                files = client.ls(self.index_config.path_without_protocol, detail=True)

        else:
            # fs.find will recursively walk directories
            # "size" is a common key for all the cloud protocols with fs
            with self.connection_config.get_client(protocol=self.index_config.protocol) as client:
                found = client.find(
                    self.index_config.path_without_protocol,
                    detail=True,
                )
                files = found.values()
        filtered_files = [
            file for file in files if file.get("size") > 0 and file.get("type") == "file"
        ]

        if self.index_config.sample_n_files:
            filtered_files = self.sample_n_files(filtered_files, self.index_config.sample_n_files)

        return filtered_files

    def sample_n_files(self, files: list[dict[str, Any]], n) -> list[dict[str, Any]]:
        if len(files) <= n:
            logger.warning(
                f"number of files to be sampled={n} is not smaller than the number"
                f" of files found ({len(files)}). Returning all of the files as the"
                " sample."
            )
            return files

        return random.sample(files, n)

    def get_metadata(self, file_info: dict) -> FileDataSourceMetadata:
        raise NotImplementedError()

    def get_path(self, file_info: dict) -> str:
        return file_info["name"]

    def sterilize_info(self, file_data: dict) -> dict:
        return sterilize_dict(data=file_data)

    def create_init_file_data(self, remote_filepath: Optional[str] = None) -> FileData:
        # Create initial file data that requires no network calls and is constructed purely
        # with information that exists in the config
        remote_filepath = remote_filepath or self.index_config.remote_url
        path_without_protocol = remote_filepath.split("://")[1]
        rel_path = remote_filepath.replace(path_without_protocol, "").lstrip("/")
        return FileData(
            identifier=str(uuid5(NAMESPACE_DNS, remote_filepath)),
            connector_type=self.connector_type,
            display_name=remote_filepath,
            source_identifiers=SourceIdentifiers(
                filename=Path(remote_filepath).name,
                rel_path=rel_path or None,
                fullpath=remote_filepath,
            ),
            metadata=FileDataSourceMetadata(url=remote_filepath),
        )

    def hydrate_file_data(self, init_file_data: FileData):
        # Get file info
        with self.connection_config.get_client(protocol=self.index_config.protocol) as client:
            files = client.ls(self.index_config.path_without_protocol, detail=True)
        filtered_files = [
            file for file in files if file.get("size") > 0 and file.get("type") == "file"
        ]
        if not filtered_files:
            raise ValueError(f"{init_file_data} did not reference any valid file")
        if len(filtered_files) > 1:
            raise ValueError(f"{init_file_data} referenced more than one file")
        file_info = filtered_files[0]
        init_file_data.additional_metadata = self.get_metadata(file_info=file_info)

    def run(self, **kwargs: Any) -> Generator[FileData, None, None]:
        files = self.get_file_info()
        for file_info in files:
            file_path = self.get_path(file_info=file_info)
            # Note: we remove any remaining leading slashes (Box introduces these)
            # to get a valid relative path
            rel_path = file_path.replace(self.index_config.path_without_protocol, "").lstrip("/")

            additional_metadata = self.sterilize_info(file_data=file_info)
            additional_metadata["original_file_path"] = file_path
            yield FileData(
                identifier=str(uuid5(NAMESPACE_DNS, file_path)),
                connector_type=self.connector_type,
                source_identifiers=SourceIdentifiers(
                    filename=Path(file_path).name,
                    rel_path=rel_path or None,
                    fullpath=file_path,
                ),
                metadata=self.get_metadata(file_info=file_info),
                additional_metadata=additional_metadata,
                display_name=file_path,
            )


class FsspecDownloaderConfig(DownloaderConfig):
    pass


FsspecDownloaderConfigT = TypeVar("FsspecDownloaderConfigT", bound=FsspecDownloaderConfig)


@dataclass
class FsspecDownloader(Downloader):
    protocol: str
    connection_config: FsspecConnectionConfigT
    connector_type: str = CONNECTOR_TYPE
    download_config: Optional[FsspecDownloaderConfigT] = field(
        default_factory=lambda: FsspecDownloaderConfig()
    )

    def is_async(self) -> bool:
        with self.connection_config.get_client(protocol=self.protocol) as client:
            return client.async_impl

    def handle_directory_download(self, lpath: Path) -> None:
        # If the object's name contains certain characters (i.e. '?'), it
        # gets downloaded into a new directory of the same name. This
        # reconciles that with what is expected, which is to download it
        # as a file that is not within a directory.
        if not lpath.is_dir():
            return
        desired_name = lpath.name
        files_in_dir = [file for file in lpath.iterdir() if file.is_file()]
        if not files_in_dir:
            raise ValueError(f"no files in {lpath}")
        if len(files_in_dir) > 1:
            raise ValueError(
                "Multiple files in {}: {}".format(lpath, ", ".join([str(f) for f in files_in_dir]))
            )
        file = files_in_dir[0]
        with tempfile.TemporaryDirectory() as temp_dir:
            temp_location = os.path.join(temp_dir, desired_name)
            shutil.copyfile(src=file, dst=temp_location)
            shutil.rmtree(lpath)
            shutil.move(src=temp_location, dst=lpath)

    def wrap_error(self, e: Exception) -> Exception:
        return self.connection_config.wrap_error(e=e)

    def run(self, file_data: FileData, **kwargs: Any) -> DownloadResponse:
        download_path = self.get_download_path(file_data=file_data)
        download_path.parent.mkdir(parents=True, exist_ok=True)
        try:
            rpath = file_data.additional_metadata["original_file_path"]
            with self.connection_config.get_client(protocol=self.protocol) as client:
                client.get(rpath=rpath, lpath=download_path.as_posix())
            self.handle_directory_download(lpath=download_path)
        except Exception as e:
            raise self.wrap_error(e=e)
        return self.generate_download_response(file_data=file_data, download_path=download_path)

    async def async_run(self, file_data: FileData, **kwargs: Any) -> DownloadResponse:
        download_path = self.get_download_path(file_data=file_data)
        download_path.parent.mkdir(parents=True, exist_ok=True)
        try:
            rpath = file_data.additional_metadata["original_file_path"]
            with self.connection_config.get_client(protocol=self.protocol) as client:
                await client.get(rpath=rpath, lpath=download_path.as_posix())
            self.handle_directory_download(lpath=download_path)
        except Exception as e:
            raise self.wrap_error(e=e)
        return self.generate_download_response(file_data=file_data, download_path=download_path)


class FsspecUploaderConfig(FileConfig, UploaderConfig):
    pass


FsspecUploaderConfigT = TypeVar("FsspecUploaderConfigT", bound=FsspecUploaderConfig)


@dataclass
class FsspecUploader(Uploader):
    connector_type: str = CONNECTOR_TYPE
    upload_config: FsspecUploaderConfigT = field(default=None)
    connection_config: FsspecConnectionConfigT

    def is_async(self) -> bool:
        with self.connection_config.get_client(protocol=self.upload_config.protocol) as client:
            return client.async_impl

    @property
    def fs(self) -> "AbstractFileSystem":
        from fsspec import get_filesystem_class

        fs_kwargs = self.connection_config.get_access_config() if self.connection_config else {}
        return get_filesystem_class(self.upload_config.protocol)(
            **fs_kwargs,
        )

    def __post_init__(self):
        # TODO once python3.9 no longer supported and kw_only is allowed in dataclasses, remove:
        if not self.upload_config:
            raise TypeError(
                f"{self.__class__.__name__}.__init__() "
                f"missing 1 required positional argument: 'upload_config'"
            )

    def wrap_error(self, e: Exception) -> Exception:
        return self.connection_config.wrap_error(e=e)

    def precheck(self) -> None:
        from fsspec import get_filesystem_class

        try:
            fs = get_filesystem_class(self.upload_config.protocol)(
                **self.connection_config.get_access_config(),
            )
            upload_path = Path(self.upload_config.path_without_protocol) / "_empty"
            fs.write_bytes(path=upload_path.as_posix(), value=b"")
        except Exception as e:
            raise self.wrap_error(e=e)

    def get_upload_path(self, file_data: FileData) -> Path:
        upload_path = (
            Path(self.upload_config.path_without_protocol)
            / file_data.source_identifiers.relative_path
        )
        updated_upload_path = upload_path.parent / f"{upload_path.name}.json"
        return updated_upload_path

    def run(self, path: Path, file_data: FileData, **kwargs: Any) -> None:
        path_str = str(path.resolve())
        upload_path = self.get_upload_path(file_data=file_data)
        logger.debug(f"writing local file {path_str} to {upload_path}")
        with self.connection_config.get_client(protocol=self.upload_config.protocol) as client:
            client.upload(lpath=path_str, rpath=upload_path.as_posix())

    async def run_async(self, path: Path, file_data: FileData, **kwargs: Any) -> None:
        upload_path = self.get_upload_path(file_data=file_data)
        path_str = str(path.resolve())
        # Odd that fsspec doesn't run exists() as async even when client support async
        logger.debug(f"writing local file {path_str} to {upload_path}")
        with self.connection_config.get_client(protocol=self.upload_config.protocol) as client:
            client.upload(lpath=path_str, rpath=upload_path.as_posix())



import os
import traceback
from dataclasses import dataclass, field
from multiprocessing import Process, Queue
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional
from urllib.parse import urlparse

from pydantic import Field, Secret

from unstructured_ingest.error import DestinationConnectionError
from unstructured_ingest.utils.data_prep import get_data_df, get_json_data
from unstructured_ingest.utils.dep_check import requires_dependencies
from unstructured_ingest.utils.table import convert_to_pandas_dataframe
from unstructured_ingest.v2.interfaces import (
    AccessConfig,
    ConnectionConfig,
    Uploader,
    UploaderConfig,
    UploadStager,
    UploadStagerConfig,
)
from unstructured_ingest.v2.logger import logger
from unstructured_ingest.v2.processes.connector_registry import DestinationRegistryEntry
from unstructured_ingest.v2.types.file_data import FileData

CONNECTOR_TYPE = "delta_table"

if TYPE_CHECKING:
    from pandas import DataFrame


@requires_dependencies(["deltalake"], extras="delta-table")
def write_deltalake_with_error_handling(queue, **kwargs):
    from deltalake.writer import write_deltalake

    try:
        write_deltalake(**kwargs)
    except Exception:
        queue.put(traceback.format_exc())


class DeltaTableAccessConfig(AccessConfig):
    aws_access_key_id: Optional[str] = Field(default=None, description="AWS Access Key Id")
    aws_secret_access_key: Optional[str] = Field(default=None, description="AWS Secret Access Key")


class DeltaTableConnectionConfig(ConnectionConfig):
    access_config: Secret[DeltaTableAccessConfig] = Field(
        default=DeltaTableAccessConfig(), validate_default=True
    )
    aws_region: Optional[str] = Field(default=None, description="AWS Region")
    table_uri: str = Field(
        default=None,
        description=(
            "Local path or path to the target folder in the S3 bucket, "
            "formatted as s3://my-bucket/my-folder/"
        ),
    )

    def update_storage_options(self, storage_options: dict) -> None:
        secrets = self.access_config.get_secret_value()
        if self.aws_region and secrets.aws_access_key_id and secrets.aws_secret_access_key:
            storage_options["AWS_REGION"] = self.aws_region
            storage_options["AWS_ACCESS_KEY_ID"] = secrets.aws_access_key_id
            storage_options["AWS_SECRET_ACCESS_KEY"] = secrets.aws_secret_access_key
            # Delta-rs doesn't support concurrent S3 writes without external locks (DynamoDB).
            # This flag allows single-writer uploads to S3 without using locks, according to:
            # https://delta-io.github.io/delta-rs/usage/writing/writing-to-s3-with-locking-provider/
            storage_options["AWS_S3_ALLOW_UNSAFE_RENAME"] = "true"


class DeltaTableUploadStagerConfig(UploadStagerConfig):
    pass


@dataclass
class DeltaTableUploadStager(UploadStager):
    upload_stager_config: DeltaTableUploadStagerConfig = field(
        default_factory=lambda: DeltaTableUploadStagerConfig()
    )

    def run(
        self,
        elements_filepath: Path,
        output_dir: Path,
        output_filename: str,
        **kwargs: Any,
    ) -> Path:
        elements_contents = get_json_data(elements_filepath)
        output_path = Path(output_dir) / Path(f"{output_filename}.parquet")

        df = convert_to_pandas_dataframe(elements_dict=elements_contents)
        df = df.dropna(axis=1, how="all")
        df.to_parquet(output_path)

        return output_path


class DeltaTableUploaderConfig(UploaderConfig):
    pass


@dataclass
class DeltaTableUploader(Uploader):
    upload_config: DeltaTableUploaderConfig
    connection_config: DeltaTableConnectionConfig
    connector_type: str = CONNECTOR_TYPE

    @requires_dependencies(["boto3"], extras="delta-table")
    def precheck(self):
        secrets = self.connection_config.access_config.get_secret_value()
        if (
            self.connection_config.aws_region
            and secrets.aws_access_key_id
            and secrets.aws_secret_access_key
        ):
            from boto3 import client

            url = urlparse(self.connection_config.table_uri)
            bucket_name = url.netloc
            dir_path = url.path.lstrip("/")

            try:
                s3_client = client(
                    "s3",
                    aws_access_key_id=secrets.aws_access_key_id,
                    aws_secret_access_key=secrets.aws_secret_access_key,
                )
                s3_client.put_object(Bucket=bucket_name, Key=dir_path, Body=b"")

                response = s3_client.get_bucket_location(Bucket=bucket_name)

                if self.connection_config.aws_region != response.get("LocationConstraint"):
                    raise ValueError("Wrong AWS Region was provided.")

            except Exception as e:
                logger.error(f"failed to validate connection: {e}", exc_info=True)
                raise DestinationConnectionError(f"failed to validate connection: {e}")

    def upload_dataframe(self, df: "DataFrame", file_data: FileData) -> None:
        updated_upload_path = os.path.join(
            self.connection_config.table_uri, file_data.source_identifiers.relative_path
        )
        logger.info(
            f"writing {len(df)} rows to destination table "
            f"at {updated_upload_path}\ndtypes: {df.dtypes}",
        )
        storage_options = {}
        self.connection_config.update_storage_options(storage_options=storage_options)

        writer_kwargs = {
            "table_or_uri": updated_upload_path,
            "data": df,
            "mode": "overwrite",
            "schema_mode": "merge",
            "storage_options": storage_options,
        }
        queue = Queue()
        # NOTE: deltalake writer on Linux sometimes can finish but still trigger a SIGABRT and cause
        # ingest to fail, even though all tasks are completed normally. Putting the writer into a
        # process mitigates this issue by ensuring python interpreter waits properly for deltalake's
        # rust backend to finish
        writer = Process(
            target=write_deltalake_with_error_handling,
            kwargs={"queue": queue, **writer_kwargs},
        )
        writer.start()
        writer.join()

        # Check if the queue has any exception message
        if not queue.empty():
            error_message = queue.get()
            logger.error(f"Exception occurred in write_deltalake: {error_message}")
            raise RuntimeError(f"Error in write_deltalake: {error_message}")

    @requires_dependencies(["pandas"], extras="delta-table")
    def run_data(self, data: list[dict], file_data: FileData, **kwargs: Any) -> None:
        import pandas as pd

        df = pd.DataFrame(data=data)
        self.upload_dataframe(df=df, file_data=file_data)

    def run(self, path: Path, file_data: FileData, **kwargs: Any) -> None:
        df = get_data_df(path)
        self.upload_dataframe(df=df, file_data=file_data)


delta_table_destination_entry = DestinationRegistryEntry(
    connection_config=DeltaTableConnectionConfig,
    uploader=DeltaTableUploader,
    uploader_config=DeltaTableUploaderConfig,
    upload_stager=DeltaTableUploadStager,
    upload_stager_config=DeltaTableUploadStagerConfig,
)



import os

import pytest

from test.integration.connectors.utils.constants import BLOB_STORAGE_TAG, SOURCE_TAG
from test.integration.connectors.utils.validation.source import (
    SourceValidationConfigs,
    source_connector_validation,
)
from test.integration.utils import requires_env
from unstructured_ingest.v2.processes.connectors.sharepoint import (
    CONNECTOR_TYPE,
    SharepointAccessConfig,
    SharepointConnectionConfig,
    SharepointDownloader,
    SharepointDownloaderConfig,
    SharepointIndexer,
    SharepointIndexerConfig,
)


def sharepoint_config():
    class SharepointTestConfig:
        def __init__(self):
            self.client_id = os.environ["SHAREPOINT_CLIENT_ID"]
            self.client_cred = os.environ["SHAREPOINT_CRED"]
            self.user_pname = os.environ["MS_USER_PNAME"]
            self.tenant = os.environ["MS_TENANT_ID"]

    return SharepointTestConfig()


@pytest.mark.asyncio
@pytest.mark.tags(CONNECTOR_TYPE, SOURCE_TAG, BLOB_STORAGE_TAG)
@requires_env("SHAREPOINT_CLIENT_ID", "SHAREPOINT_CRED", "MS_TENANT_ID", "MS_USER_PNAME")
async def test_sharepoint_source(temp_dir):
    site = "https://unstructuredio.sharepoint.com/sites/utic-platform-test-source"
    config = sharepoint_config()

    # Create connection and indexer configurations
    access_config = SharepointAccessConfig(client_cred=config.client_cred)
    connection_config = SharepointConnectionConfig(
        client_id=config.client_id,
        site=site,
        tenant=config.tenant,
        user_pname=config.user_pname,
        access_config=access_config,
    )
    index_config = SharepointIndexerConfig(recursive=True)

    download_config = SharepointDownloaderConfig(download_dir=temp_dir)

    # Instantiate indexer and downloader
    indexer = SharepointIndexer(
        connection_config=connection_config,
        index_config=index_config,
    )
    downloader = SharepointDownloader(
        connection_config=connection_config,
        download_config=download_config,
    )

    # Run the source connector validation
    await source_connector_validation(
        indexer=indexer,
        downloader=downloader,
        configs=SourceValidationConfigs(
            test_id="sharepoint1",
            expected_num_files=4,
            validate_downloaded_files=True,
            exclude_fields_extend=[
                "metadata.date_created",
                "metadata.date_modified",
                "additional_metadata.LastModified",
                "additional_metadata.@microsoft.graph.downloadUrl",
            ],
        ),
    )


@pytest.mark.asyncio
@pytest.mark.tags(CONNECTOR_TYPE, SOURCE_TAG, BLOB_STORAGE_TAG)
@requires_env("SHAREPOINT_CLIENT_ID", "SHAREPOINT_CRED", "MS_TENANT_ID", "MS_USER_PNAME")
async def test_sharepoint_source_with_path(temp_dir):
    site = "https://unstructuredio.sharepoint.com/sites/utic-platform-test-source"
    config = sharepoint_config()

    # Create connection and indexer configurations
    access_config = SharepointAccessConfig(client_cred=config.client_cred)
    connection_config = SharepointConnectionConfig(
        client_id=config.client_id,
        site=site,
        tenant=config.tenant,
        user_pname=config.user_pname,
        access_config=access_config,
    )
    index_config = SharepointIndexerConfig(recursive=True, path="Folder1")

    download_config = SharepointDownloaderConfig(download_dir=temp_dir)

    # Instantiate indexer and downloader
    indexer = SharepointIndexer(
        connection_config=connection_config,
        index_config=index_config,
    )
    downloader = SharepointDownloader(
        connection_config=connection_config,
        download_config=download_config,
    )

    # Run the source connector validation
    await source_connector_validation(
        indexer=indexer,
        downloader=downloader,
        configs=SourceValidationConfigs(
            test_id="sharepoint2",
            expected_num_files=2,
            validate_downloaded_files=True,
            exclude_fields_extend=[
                "metadata.date_created",
                "metadata.date_modified",
                "additional_metadata.LastModified",
                "additional_metadata.@microsoft.graph.downloadUrl",
            ],
        ),
    )


@pytest.mark.asyncio
@pytest.mark.tags(CONNECTOR_TYPE, SOURCE_TAG, BLOB_STORAGE_TAG)
@requires_env("SHAREPOINT_CLIENT_ID", "SHAREPOINT_CRED", "MS_TENANT_ID", "MS_USER_PNAME")
async def test_sharepoint_root_with_path(temp_dir):
    site = "https://unstructuredio.sharepoint.com/"
    config = sharepoint_config()

    # Create connection and indexer configurations
    access_config = SharepointAccessConfig(client_cred=config.client_cred)
    connection_config = SharepointConnectionConfig(
        client_id=config.client_id,
        site=site,
        tenant=config.tenant,
        user_pname=config.user_pname,
        access_config=access_config,
    )
    index_config = SharepointIndexerConfig(recursive=True, path="e2e-test-folder")

    download_config = SharepointDownloaderConfig(download_dir=temp_dir)

    # Instantiate indexer and downloader
    indexer = SharepointIndexer(
        connection_config=connection_config,
        index_config=index_config,
    )
    downloader = SharepointDownloader(
        connection_config=connection_config,
        download_config=download_config,
    )

    # Run the source connector validation
    await source_connector_validation(
        indexer=indexer,
        downloader=downloader,
        configs=SourceValidationConfigs(
            test_id="sharepoint3",
            expected_num_files=1,
            validate_downloaded_files=True,
            exclude_fields_extend=[
                "metadata.date_created",
                "metadata.date_modified",
                "additional_metadata.LastModified",
                "additional_metadata.@microsoft.graph.downloadUrl",
            ],
        ),
    )


@pytest.mark.asyncio
@pytest.mark.tags(CONNECTOR_TYPE, SOURCE_TAG, BLOB_STORAGE_TAG)
@requires_env("SHAREPOINT_CLIENT_ID", "SHAREPOINT_CRED", "MS_TENANT_ID", "MS_USER_PNAME")
async def test_sharepoint_shared_documents(temp_dir):
    site = "https://unstructuredio.sharepoint.com/sites/utic-platform-test-source"
    config = sharepoint_config()

    # Create connection and indexer configurations
    access_config = SharepointAccessConfig(client_cred=config.client_cred)
    connection_config = SharepointConnectionConfig(
        client_id=config.client_id,
        site=site,
        tenant=config.tenant,
        user_pname=config.user_pname,
        access_config=access_config,
    )
    index_config = SharepointIndexerConfig(recursive=True, path="Shared Documents")

    download_config = SharepointDownloaderConfig(download_dir=temp_dir)

    # Instantiate indexer and downloader
    indexer = SharepointIndexer(
        connection_config=connection_config,
        index_config=index_config,
    )
    downloader = SharepointDownloader(
        connection_config=connection_config,
        download_config=download_config,
    )

    # Run the source connector validation
    await source_connector_validation(
        indexer=indexer,
        downloader=downloader,
        configs=SourceValidationConfigs(
            test_id="sharepoint4",
            expected_num_files=4,
            validate_downloaded_files=True,
            exclude_fields_extend=[
                "metadata.date_created",
                "metadata.date_modified",
                "additional_metadata.LastModified",
                "additional_metadata.@microsoft.graph.downloadUrl",
            ],
        ),
    )



import hashlib
import typing as t
from dataclasses import dataclass

from unstructured_ingest.interfaces import BaseSourceConnector
from unstructured_ingest.logger import logger
from unstructured_ingest.runner.base_runner import Runner
from unstructured_ingest.runner.utils import update_download_dir_hash

if t.TYPE_CHECKING:
    from unstructured_ingest.connector.onedrive import SimpleOneDriveConfig


@dataclass
class OneDriveRunner(Runner):
    connector_config: "SimpleOneDriveConfig"

    def update_read_config(self):
        hashed_dir_name = hashlib.sha256(
            f"{self.connector_config.tenant}_{self.connector_config.user_pname}".encode("utf-8"),
        )

        self.read_config.download_dir = update_download_dir_hash(
            connector_name="onedrive",
            read_config=self.read_config,
            hashed_dir_name=hashed_dir_name,
            logger=logger,
        )

    def get_source_connector_cls(self) -> t.Type[BaseSourceConnector]:
        from unstructured_ingest.connector.onedrive import (
            OneDriveSourceConnector,
        )

        return OneDriveSourceConnector



import typing as t
from dataclasses import dataclass

import click

from unstructured_ingest.cli.base.src import BaseSrcCmd
from unstructured_ingest.cli.interfaces import CliConfig, DelimitedString
from unstructured_ingest.connector.mongodb import SimpleMongoDBConfig
from unstructured_ingest.interfaces import WriteConfig

CMD_NAME = "mongodb"


@dataclass
class MongoDBCliConfig(SimpleMongoDBConfig, CliConfig):
    @staticmethod
    def get_cli_options() -> t.List[click.Option]:
        options = [
            click.Option(
                ["--uri"],
                help="URI to user when connecting",
            ),
            click.Option(
                ["--host"],
                type=DelimitedString(),
                help="hostname or IP address or Unix domain socket path of a single mongod or "
                "mongos instance to connect to, or a list of hostnames",
            ),
            click.Option(["--port"], type=int, default=27017),
            click.Option(
                ["--database"], type=str, required=True, help="database name to connect to"
            ),
            click.Option(
                ["--collection"], required=True, type=str, help="collection name to connect to"
            ),
        ]
        return options


@dataclass
class MongoDBReadConfig(SimpleMongoDBConfig, CliConfig):
    @staticmethod
    def get_cli_options() -> t.List[click.Option]:
        options = [
            click.Option(
                ["--batch-size"],
                default=100,
                type=click.IntRange(0),
                help="how many records to read at a time per process",
            ),
        ]
        return options


def get_base_src_cmd() -> BaseSrcCmd:
    cmd_cls = BaseSrcCmd(
        cmd_name=CMD_NAME,
        cli_config=MongoDBCliConfig,
        additional_cli_options=[MongoDBReadConfig],
    )
    return cmd_cls


def get_base_dest_cmd():
    from unstructured_ingest.cli.base.dest import BaseDestCmd

    cmd_cls = BaseDestCmd(
        cmd_name=CMD_NAME,
        cli_config=MongoDBCliConfig,
        write_config=WriteConfig,
    )
    return cmd_cls



import os

import pytest

from test.integration.connectors.utils.constants import SOURCE_TAG, UNCATEGORIZED_TAG
from test.integration.connectors.utils.validation.source import (
    SourceValidationConfigs,
    get_all_file_data,
    run_all_validations,
    update_fixtures,
)
from unstructured_ingest.v2.interfaces import Downloader, Indexer
from unstructured_ingest.v2.processes.connectors.notion.connector import (
    CONNECTOR_TYPE,
    NotionAccessConfig,
    NotionConnectionConfig,
    NotionDownloader,
    NotionDownloaderConfig,
    NotionIndexer,
    NotionIndexerConfig,
)


@pytest.mark.tags(SOURCE_TAG, CONNECTOR_TYPE, UNCATEGORIZED_TAG)
def test_notion_source_database(temp_dir):
    # Retrieve environment variables
    notion_api_key = os.environ["NOTION_API_KEY"]

    # Create connection and indexer configurations
    access_config = NotionAccessConfig(notion_api_key=notion_api_key)
    connection_config = NotionConnectionConfig(
        access_config=access_config,
    )
    index_config = NotionIndexerConfig(
        database_ids=["1722c3765a0a8082b382ebc2c62d3f4c"], recursive=False
    )

    download_config = NotionDownloaderConfig(download_dir=temp_dir)

    # Instantiate indexer and downloader
    indexer = NotionIndexer(
        connection_config=connection_config,
        index_config=index_config,
    )
    downloader = NotionDownloader(
        connection_config=connection_config,
        download_config=download_config,
    )

    # Run the source connector validation
    source_connector_validation(
        indexer=indexer,
        downloader=downloader,
        configs=SourceValidationConfigs(
            test_id="notion_database",
            expected_num_files=1,
            validate_downloaded_files=True,
            exclude_fields_extend=["metadata.date_created", "metadata.date_modified"],
        ),
    )


@pytest.mark.tags(SOURCE_TAG, CONNECTOR_TYPE, UNCATEGORIZED_TAG)
def test_notion_source_page(temp_dir):
    # Retrieve environment variables
    notion_api_key = os.environ["NOTION_API_KEY"]

    # Create connection and indexer configurations
    access_config = NotionAccessConfig(notion_api_key=notion_api_key)
    connection_config = NotionConnectionConfig(
        access_config=access_config,
    )
    index_config = NotionIndexerConfig(
        page_ids=["1572c3765a0a806299f0dd6999f9e4c7"], recursive=False
    )

    download_config = NotionDownloaderConfig(download_dir=temp_dir)

    # Instantiate indexer and downloader
    indexer = NotionIndexer(
        connection_config=connection_config,
        index_config=index_config,
    )
    downloader = NotionDownloader(
        connection_config=connection_config,
        download_config=download_config,
    )

    # Run the source connector validation
    source_connector_validation(
        indexer=indexer,
        downloader=downloader,
        configs=SourceValidationConfigs(
            test_id="notion_page",
            expected_num_files=1,
            validate_downloaded_files=True,
            exclude_fields_extend=["metadata.date_created", "metadata.date_modified"],
        ),
    )


@pytest.mark.tags(SOURCE_TAG, CONNECTOR_TYPE, UNCATEGORIZED_TAG)
def source_connector_validation(
    indexer: Indexer,
    downloader: Downloader,
    configs: SourceValidationConfigs,
    overwrite_fixtures: bool = os.getenv("OVERWRITE_FIXTURES", "False").lower() == "true",
) -> None:
    # Run common validations on the process of running a source connector, supporting dynamic
    # validators that get passed in along with comparisons on the saved expected values.
    # If overwrite_fixtures is st to True, will ignore all validators but instead overwrite the
    # expected values with what gets generated by this test.
    all_predownload_file_data = []
    all_postdownload_file_data = []
    indexer.precheck()
    download_dir = downloader.download_config.download_dir
    test_output_dir = configs.test_output_dir()

    for file_data in indexer.run():
        assert file_data
        predownload_file_data = file_data.model_copy(deep=True)
        all_predownload_file_data.append(predownload_file_data)
        resp = downloader.run(file_data=file_data)
        if isinstance(resp, list):
            for r in resp:
                postdownload_file_data = r["file_data"].model_copy(deep=True)
                all_postdownload_file_data.append(postdownload_file_data)
        else:
            postdownload_file_data = resp["file_data"].model_copy(deep=True)
            all_postdownload_file_data.append(postdownload_file_data)

    if not overwrite_fixtures:
        print("Running validation")
        run_all_validations(
            configs=configs,
            predownload_file_data=all_predownload_file_data,
            postdownload_file_data=all_postdownload_file_data,
            download_dir=download_dir,
            test_output_dir=test_output_dir,
        )
    else:
        print("Running fixtures update")
        update_fixtures(
            output_dir=test_output_dir,
            download_dir=download_dir,
            all_file_data=get_all_file_data(
                all_predownload_file_data=all_predownload_file_data,
                all_postdownload_file_data=all_postdownload_file_data,
            ),
            save_downloads=configs.validate_downloaded_files,
            save_filedata=configs.validate_file_data,
        )



import random
from typing import Any, Type

from faker import Faker

fake = Faker()

type_to_random_value_map = {
    str: fake.sentence,
    int: fake.random_int,
    float: fake.random_digit,
    bool: fake.boolean,
}
type_to_random_value_map_key = type_to_random_value_map.copy()
type_to_random_value_map_key[str] = fake.word


def generate_random_dictionary(key_type: Type = str, value_type: Type = str) -> dict:
    d = {}
    num_keys = random.randint(1, 3)
    for i in range(num_keys):
        key = type_to_random_value_map_key[key_type]()
        current_value_type = value_type
        if current_value_type == Any:
            current_value_type = random.choice(list(type_to_random_value_map.keys()) + [dict])
        value = (
            generate_random_dictionary(key_type=key_type, value_type=value_type)
            if current_value_type is dict
            else type_to_random_value_map[current_value_type]()
        )
        d[key] = value
    return d



import os

import pytest

from test.integration.connectors.utils.constants import SOURCE_TAG, UNCATEGORIZED_TAG
from test.integration.connectors.utils.validation.source import (
    SourceValidationConfigs,
    source_connector_validation,
)
from test.integration.utils import requires_env
from unstructured_ingest.v2.processes.connectors.jira import (
    CONNECTOR_TYPE,
    JiraAccessConfig,
    JiraConnectionConfig,
    JiraDownloader,
    JiraDownloaderConfig,
    JiraIndexer,
    JiraIndexerConfig,
)


@pytest.mark.asyncio
@pytest.mark.tags(CONNECTOR_TYPE, SOURCE_TAG, UNCATEGORIZED_TAG)
@requires_env("JIRA_INGEST_USER_EMAIL", "JIRA_INGEST_API_TOKEN")
async def test_jira_source(temp_dir):
    # Retrieve environment variables
    jira_url = os.environ.get(
        "JIRA_INGEST_URL", "https://unstructured-jira-connector-test.atlassian.net"
    )
    user_email = os.environ["JIRA_INGEST_USER_EMAIL"]
    api_token = os.environ["JIRA_INGEST_API_TOKEN"]
    projects = ["JCTP1"]
    boards = ["3"]
    issues = ["JCTP2-1", "JCTP2-2", "JCTP2-3"]

    # Create connection and indexer configurations
    access_config = JiraAccessConfig(password=api_token)
    connection_config = JiraConnectionConfig(
        url=jira_url,
        username=user_email,
        access_config=access_config,
    )
    index_config = JiraIndexerConfig(projects=projects, boards=boards, issues=issues)

    download_config = JiraDownloaderConfig(download_dir=temp_dir)

    # Instantiate indexer and downloader
    indexer = JiraIndexer(
        connection_config=connection_config,
        index_config=index_config,
    )
    downloader = JiraDownloader(
        connection_config=connection_config,
        download_config=download_config,
    )

    # Run the source connector validation
    await source_connector_validation(
        indexer=indexer,
        downloader=downloader,
        configs=SourceValidationConfigs(
            test_id="jira",
            expected_num_files=8,
            validate_file_data=True,
            validate_downloaded_files=True,
        ),
    )



from abc import ABC
from typing import Any, AsyncGenerator, Generator, Optional, TypeVar

from pydantic import BaseModel

from unstructured_ingest.v2.interfaces.connector import BaseConnector
from unstructured_ingest.v2.interfaces.process import BaseProcess
from unstructured_ingest.v2.types.file_data import FileData


class IndexerConfig(BaseModel):
    pass


IndexerConfigT = TypeVar("IndexerConfigT", bound=IndexerConfig)


class Indexer(BaseProcess, BaseConnector, ABC):
    connector_type: str
    index_config: Optional[IndexerConfigT] = None

    def is_async(self) -> bool:
        return False

    def run(self, **kwargs: Any) -> Generator[FileData, None, None]:
        raise NotImplementedError()

    async def run_async(self, **kwargs: Any) -> AsyncGenerator[FileData, None]:
        raise NotImplementedError()



import math
from collections import abc
from contextlib import contextmanager
from dataclasses import dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Generator, List, Optional, Union

from pydantic import Field, Secret

from unstructured_ingest.error import SourceConnectionError
from unstructured_ingest.utils.dep_check import requires_dependencies
from unstructured_ingest.v2.interfaces import (
    AccessConfig,
    ConnectionConfig,
    Downloader,
    DownloaderConfig,
    DownloadResponse,
    Indexer,
    IndexerConfig,
)
from unstructured_ingest.v2.logger import logger
from unstructured_ingest.v2.processes.connector_registry import (
    SourceRegistryEntry,
)
from unstructured_ingest.v2.types.file_data import (
    FileData,
    FileDataSourceMetadata,
    SourceIdentifiers,
)

if TYPE_CHECKING:
    from atlassian import Jira

CONNECTOR_TYPE = "jira"

DEFAULT_C_SEP = " " * 5
DEFAULT_R_SEP = "\n"


@dataclass
class JiraIssueMetadata:
    id: str
    key: str
    board_id: Optional[str] = None

    @property
    def project_id(self) -> str:
        return self.key.split("-")[0]

    def to_dict(self) -> Dict[str, Union[str, None]]:
        return {
            "id": self.id,
            "key": self.key,
            "board_id": self.board_id,
            "project_id": self.project_id,
        }


class FieldGetter(dict):
    def __getitem__(self, key):
        value = super().__getitem__(key) if key in self else None
        if value is None:
            value = FieldGetter({})
        return value


def nested_object_to_field_getter(obj: dict) -> Union[FieldGetter, dict]:
    if isinstance(obj, abc.Mapping):
        new_object = {}
        for k, v in obj.items():
            if isinstance(v, abc.Mapping):
                new_object[k] = FieldGetter(nested_object_to_field_getter(v))
            else:
                new_object[k] = v
        return FieldGetter(new_object)
    else:
        return obj


def issues_fetcher_wrapper(func, results_key="results", number_of_issues_to_fetch: int = 100):
    """
    A decorator function that wraps around a function to fetch issues from Jira API in a paginated
    manner. This is required because the Jira API has a limit of 100 issues per request.

    Args:
        func (callable): The function to be wrapped. This function should accept `limit` and `start`
        as keyword arguments.
        results_key (str, optional): The key in the response dictionary that contains the list of
        results. Defaults to "results".
        number_of_issues_to_fetch (int, optional): The total number of issues to fetch. Defaults to
        100.

    Returns:
        list: A list of all fetched issues.

    Raises:
        KeyError: If the response dictionary does not contain the specified `results_key`.
        TypeError: If the response type from the Jira API is neither list nor dict.
    """

    def wrapper(*args, **kwargs) -> list:
        kwargs["limit"] = min(100, number_of_issues_to_fetch)
        kwargs["start"] = kwargs.get("start", 0)

        all_results = []
        num_iterations = math.ceil(number_of_issues_to_fetch / kwargs["limit"])

        for _ in range(num_iterations):
            response = func(*args, **kwargs)
            if isinstance(response, list):
                all_results += response
            elif isinstance(response, dict):
                if results_key not in response:
                    raise KeyError(f'Response object is missing "{results_key}" key.')
                all_results += response[results_key]
            else:
                raise TypeError(
                    f"""Unexpected response type from Jira API.
                    Response type has to be either list or dict, got: {type(response).__name__}."""
                )
            kwargs["start"] += kwargs["limit"]

        return all_results

    return wrapper


class JiraAccessConfig(AccessConfig):
    password: Optional[str] = Field(
        description="Jira password or Cloud API token",
        default=None,
    )
    token: Optional[str] = Field(
        description="Jira Personal Access Token",
        default=None,
    )


class JiraConnectionConfig(ConnectionConfig):
    url: str = Field(description="URL of the Jira instance")
    username: Optional[str] = Field(
        description="Username or email for authentication",
        default=None,
    )
    cloud: bool = Field(description="Authenticate to Jira Cloud", default=False)
    access_config: Secret[JiraAccessConfig] = Field(description="Access configuration for Jira")

    def model_post_init(self, __context):
        access_configs = self.access_config.get_secret_value()
        basic_auth = self.username and access_configs.password
        pat_auth = access_configs.token
        if self.cloud and not basic_auth:
            raise ValueError(
                "cloud authentication requires username and API token (--password), "
                "see: https://atlassian-python-api.readthedocs.io/"
            )
        if basic_auth and pat_auth:
            raise ValueError(
                "both password and token provided, only one allowed, "
                "see: https://atlassian-python-api.readthedocs.io/"
            )
        if not (basic_auth or pat_auth):
            raise ValueError(
                "no form of auth provided, see: https://atlassian-python-api.readthedocs.io/"
            )

    @requires_dependencies(["atlassian"], extras="jira")
    @contextmanager
    def get_client(self) -> Generator["Jira", None, None]:
        from atlassian import Jira

        access_configs = self.access_config.get_secret_value()
        with Jira(
            url=self.url,
            username=self.username,
            password=access_configs.password,
            token=access_configs.token,
            cloud=self.cloud,
        ) as client:
            yield client


class JiraIndexerConfig(IndexerConfig):
    projects: Optional[List[str]] = Field(None, description="List of project keys")
    boards: Optional[List[str]] = Field(None, description="List of board IDs")
    issues: Optional[List[str]] = Field(None, description="List of issue keys or IDs")


@dataclass
class JiraIndexer(Indexer):
    connection_config: JiraConnectionConfig
    index_config: JiraIndexerConfig
    connector_type: str = CONNECTOR_TYPE

    def precheck(self) -> None:
        try:
            with self.connection_config.get_client() as client:
                response = client.get_permissions("BROWSE_PROJECTS")
                permitted = response["permissions"]["BROWSE_PROJECTS"]["havePermission"]
        except Exception as e:
            logger.error(f"Failed to connect to Jira: {e}", exc_info=True)
            raise SourceConnectionError(f"Failed to connect to Jira: {e}")
        if not permitted:
            raise ValueError(
                """The provided user is not permitted to browse projects
                from the given Jira organization URL.
                Try checking username, password, token and the url arguments.""",
            )
        logger.info("Connection to Jira successful.")

    def _get_issues_within_single_project(self, project_key: str) -> List[JiraIssueMetadata]:
        with self.connection_config.get_client() as client:
            number_of_issues_to_fetch = client.get_project_issues_count(project=project_key)
            if isinstance(number_of_issues_to_fetch, dict):
                if "total" not in number_of_issues_to_fetch:
                    raise KeyError('Response object is missing "total" key.')
                number_of_issues_to_fetch = number_of_issues_to_fetch["total"]
            if not number_of_issues_to_fetch:
                logger.warning(f"No issues found in project: {project_key}. Skipping!")
                return []
            get_project_issues = issues_fetcher_wrapper(
                client.get_all_project_issues,
                results_key="issues",
                number_of_issues_to_fetch=number_of_issues_to_fetch,
            )
            issues = get_project_issues(project=project_key, fields=["key", "id"])
            logger.debug(f"Found {len(issues)} issues in project: {project_key}")
            return [JiraIssueMetadata(id=issue["id"], key=issue["key"]) for issue in issues]

    def _get_issues_within_projects(self) -> List[JiraIssueMetadata]:
        project_keys = self.index_config.projects
        if not project_keys:
            # for when a component list is provided, without any projects
            if self.index_config.boards or self.index_config.issues:
                return []
            # for when no components are provided. all projects will be ingested
            else:
                with self.connection_config.get_client() as client:
                    project_keys = [project["key"] for project in client.projects()]
        return [
            issue
            for project_key in project_keys
            for issue in self._get_issues_within_single_project(project_key)
        ]

    def _get_issues_within_single_board(self, board_id: str) -> List[JiraIssueMetadata]:
        with self.connection_config.get_client() as client:
            get_board_issues = issues_fetcher_wrapper(
                client.get_issues_for_board,
                results_key="issues",
            )
            issues = get_board_issues(board_id=board_id, fields=["key", "id"], jql=None)
            logger.debug(f"Found {len(issues)} issues in board: {board_id}")
            return [
                JiraIssueMetadata(id=issue["id"], key=issue["key"], board_id=board_id)
                for issue in issues
            ]

    def _get_issues_within_boards(self) -> List[JiraIssueMetadata]:
        if not self.index_config.boards:
            return []
        return [
            issue
            for board_id in self.index_config.boards
            for issue in self._get_issues_within_single_board(board_id)
        ]

    def _get_issues(self) -> List[JiraIssueMetadata]:
        with self.connection_config.get_client() as client:
            issues = [
                client.get_issue(issue_id_or_key=issue_key, fields=["key", "id"])
                for issue_key in self.index_config.issues or []
            ]
        return [JiraIssueMetadata(id=issue["id"], key=issue["key"]) for issue in issues]

    def get_issues(self) -> List[JiraIssueMetadata]:
        issues = [
            *self._get_issues_within_boards(),
            *self._get_issues_within_projects(),
            *self._get_issues(),
        ]
        # Select unique issues by issue 'id'.
        # Since boards issues are fetched first,
        # if there are duplicates, the board issues will be kept,
        # in order to keep issue 'board_id' information.
        seen = set()
        unique_issues: List[JiraIssueMetadata] = []
        for issue in issues:
            if issue.id not in seen:
                unique_issues.append(issue)
            seen.add(issue.id)
        return unique_issues

    def run(self, **kwargs: Any) -> Generator[FileData, None, None]:
        from time import time

        issues = self.get_issues()
        for issue in issues:
            # Build metadata
            metadata = FileDataSourceMetadata(
                date_processed=str(time()),
                record_locator=issue.to_dict(),
            )

            # Construct relative path and filename
            filename = f"{issue.id}.txt"
            relative_path = str(Path(issue.project_id) / filename)

            source_identifiers = SourceIdentifiers(
                filename=filename,
                fullpath=relative_path,
                rel_path=relative_path,
            )

            file_data = FileData(
                identifier=issue.id,
                connector_type=self.connector_type,
                metadata=metadata,
                additional_metadata=issue.to_dict(),
                source_identifiers=source_identifiers,
            )
            yield file_data


class JiraDownloaderConfig(DownloaderConfig):
    pass


@dataclass
class JiraDownloader(Downloader):
    connection_config: JiraConnectionConfig
    download_config: JiraDownloaderConfig = field(default_factory=JiraDownloaderConfig)
    connector_type: str = CONNECTOR_TYPE

    def _get_id_fields_for_issue(
        self, issue: dict, c_sep: str = DEFAULT_C_SEP, r_sep: str = DEFAULT_R_SEP
    ) -> str:
        issue_id, key = issue["id"], issue["key"]
        return f"IssueID_IssueKey:{issue_id}{c_sep}{key}{r_sep}"

    def _get_project_fields_for_issue(
        self, issue: dict, c_sep: str = DEFAULT_C_SEP, r_sep: str = DEFAULT_R_SEP
    ) -> str:
        if "project" in issue:
            return (
                f"ProjectID_Key:{issue['project']['key']}{c_sep}{issue['project']['name']}{r_sep}"
            )
        else:
            return ""

    def _get_dropdown_fields_for_issue(
        self, issue: dict, c_sep: str = DEFAULT_C_SEP, r_sep: str = DEFAULT_R_SEP
    ) -> str:
        return f"""
        IssueType:{issue["issuetype"]["name"]}
        {r_sep}
        Status:{issue["status"]["name"]}
        {r_sep}
        Priority:{issue["priority"]}
        {r_sep}
        AssigneeID_Name:{issue["assignee"]["accountId"]}{c_sep}{issue["assignee"]["displayName"]}
        {r_sep}
        ReporterAdr_Name:{issue["reporter"]["emailAddress"]}{c_sep}{issue["reporter"]["displayName"]}
        {r_sep}
        Labels:{c_sep.join(issue["labels"])}
        {r_sep}
        Components:{c_sep.join([component["name"] for component in issue["components"]])}
        {r_sep}
        """

    def _get_subtasks_for_issue(self, issue: dict) -> str:
        return ""

    def _get_text_fields_for_issue(
        self, issue: dict, c_sep: str = DEFAULT_C_SEP, r_sep: str = DEFAULT_R_SEP
    ) -> str:
        return f"""
        {issue["summary"]}
        {r_sep}
        {issue["description"]}
        {r_sep}
        {c_sep.join([attachment["self"] for attachment in issue["attachment"]])}
        {r_sep}
        """

    def _get_comments_for_issue(
        self, issue: dict, c_sep: str = DEFAULT_C_SEP, r_sep: str = DEFAULT_R_SEP
    ) -> str:
        return c_sep.join(
            [self._get_fields_for_comment(comment) for comment in issue["comment"]["comments"]],
        )

    def _get_fields_for_comment(
        self, comment, c_sep: str = DEFAULT_C_SEP, r_sep: str = DEFAULT_R_SEP
    ) -> str:
        return f"{comment['author']['displayName']}{c_sep}{comment['body']}{r_sep}"

    def form_templated_string(
        self,
        issue: dict,
        parsed_fields: Union[FieldGetter, dict],
        c_sep: str = "|||",
        r_sep: str = "\n\n\n",
    ) -> str:
        """Forms a template string via parsing the fields from the API response object on the issue
        The template string will be saved to the disk, and then will be processed by partition."""
        return r_sep.join(
            [
                self._get_id_fields_for_issue(issue),
                self._get_project_fields_for_issue(parsed_fields),
                self._get_dropdown_fields_for_issue(parsed_fields),
                self._get_subtasks_for_issue(parsed_fields),
                self._get_comments_for_issue(parsed_fields),
                self._get_text_fields_for_issue(parsed_fields),
            ],
        )

    def update_file_data(self, file_data: FileData, issue: dict) -> None:
        file_data.metadata.date_created = issue["fields"]["created"]
        file_data.metadata.date_modified = issue["fields"]["updated"]
        file_data.display_name = issue["fields"]["project"]["name"]

    def get_issue(self, issue_key: str) -> dict:
        try:
            with self.connection_config.get_client() as client:
                return client.issue(key=issue_key)
        except Exception as e:
            logger.error(f"Failed to fetch issue with key: {issue_key}: {e}", exc_info=True)
            raise SourceConnectionError(f"Failed to fetch issue with key: {issue_key}: {e}")

    def run(self, file_data: FileData, **kwargs: Any) -> DownloadResponse:
        issue_key = file_data.additional_metadata.get("key")
        if not issue_key:
            raise ValueError("Issue key not found in metadata.")
        issue = self.get_issue(issue_key)
        parsed_fields = nested_object_to_field_getter(issue["fields"])
        issue_str = self.form_templated_string(issue, parsed_fields)

        download_path = self.get_download_path(file_data)
        if download_path is None:
            raise ValueError("File data is missing source identifiers data.")
        download_path.parent.mkdir(parents=True, exist_ok=True)
        with open(download_path, "w") as f:
            f.write(issue_str)
        self.update_file_data(file_data, issue)
        return self.generate_download_response(file_data=file_data, download_path=download_path)


jira_source_entry = SourceRegistryEntry(
    connection_config=JiraConnectionConfig,
    indexer_config=JiraIndexerConfig,
    indexer=JiraIndexer,
    downloader_config=JiraDownloaderConfig,
    downloader=JiraDownloader,
)



import json
from typing import Any

from pydantic import BaseModel, Field, Secret, SecretStr
from pydantic.types import _SecretBase

from unstructured_ingest.v2.utils import serialize_base_model, serialize_base_model_json


class MockChildBaseModel(BaseModel):
    child_secret_str: SecretStr
    child_secret_float: Secret[float]
    child_not_secret_dict: dict[str, Any] = Field(default_factory=dict)


class MockBaseModel(BaseModel):
    secret_str: SecretStr
    not_secret_bool: bool
    secret_child_base: Secret[MockChildBaseModel]
    not_secret_list: list[int] = Field(default_factory=list)


model = MockBaseModel(
    secret_str="secret string",
    not_secret_bool=False,
    secret_child_base=MockChildBaseModel(
        child_secret_str="child secret string",
        child_secret_float=3.14,
        child_not_secret_dict={"key": "value"},
    ),
    not_secret_list=[1, 2, 3],
)


def test_serialize_base_model():

    serialized_dict = model.model_dump()
    assert isinstance(serialized_dict["secret_str"], _SecretBase)
    assert isinstance(serialized_dict["secret_child_base"], _SecretBase)

    serialized_dict_w_secrets = serialize_base_model(model=model)
    assert not isinstance(serialized_dict_w_secrets["secret_str"], _SecretBase)
    assert not isinstance(serialized_dict_w_secrets["secret_child_base"], _SecretBase)

    expected_dict = {
        "secret_str": "secret string",
        "not_secret_bool": False,
        "secret_child_base": {
            "child_secret_str": "child secret string",
            "child_secret_float": 3.14,
            "child_not_secret_dict": {"key": "value"},
        },
        "not_secret_list": [1, 2, 3],
    }

    assert serialized_dict_w_secrets == expected_dict


def test_serialize_base_model_json():
    serialized_json = model.model_dump_json()
    serialized_dict = json.loads(serialized_json)
    expected_dict = {
        "secret_str": "**********",
        "not_secret_bool": False,
        "secret_child_base": "**********",
        "not_secret_list": [1, 2, 3],
    }
    assert expected_dict == serialized_dict

    serialized_json_w_secrets = serialize_base_model_json(model=model)
    serialized_dict_w_secrets = json.loads(serialized_json_w_secrets)
    expected_dict_w_secrets = {
        "secret_str": "secret string",
        "not_secret_bool": False,
        "secret_child_base": {
            "child_secret_str": "child secret string",
            "child_secret_float": 3.14,
            "child_not_secret_dict": {"key": "value"},
        },
        "not_secret_list": [1, 2, 3],
    }
    assert expected_dict_w_secrets == serialized_dict_w_secrets



from __future__ import annotations

from unstructured_ingest.v2.processes.connector_registry import (
    add_destination_entry,
    add_source_entry,
)

from .azure import CONNECTOR_TYPE as AZURE_CONNECTOR_TYPE
from .azure import azure_destination_entry, azure_source_entry
from .box import CONNECTOR_TYPE as BOX_CONNECTOR_TYPE
from .box import box_destination_entry, box_source_entry
from .dropbox import CONNECTOR_TYPE as DROPBOX_CONNECTOR_TYPE
from .dropbox import dropbox_destination_entry, dropbox_source_entry
from .gcs import CONNECTOR_TYPE as GCS_CONNECTOR_TYPE
from .gcs import gcs_destination_entry, gcs_source_entry
from .s3 import CONNECTOR_TYPE as S3_CONNECTOR_TYPE
from .s3 import s3_destination_entry, s3_source_entry
from .sftp import CONNECTOR_TYPE as SFTP_CONNECTOR_TYPE
from .sftp import sftp_destination_entry, sftp_source_entry

add_source_entry(source_type=AZURE_CONNECTOR_TYPE, entry=azure_source_entry)
add_destination_entry(destination_type=AZURE_CONNECTOR_TYPE, entry=azure_destination_entry)

add_source_entry(source_type=BOX_CONNECTOR_TYPE, entry=box_source_entry)
add_destination_entry(destination_type=BOX_CONNECTOR_TYPE, entry=box_destination_entry)

add_source_entry(source_type=DROPBOX_CONNECTOR_TYPE, entry=dropbox_source_entry)
add_destination_entry(destination_type=DROPBOX_CONNECTOR_TYPE, entry=dropbox_destination_entry)

add_source_entry(source_type=GCS_CONNECTOR_TYPE, entry=gcs_source_entry)
add_destination_entry(destination_type=GCS_CONNECTOR_TYPE, entry=gcs_destination_entry)

add_source_entry(source_type=S3_CONNECTOR_TYPE, entry=s3_source_entry)
add_destination_entry(destination_type=S3_CONNECTOR_TYPE, entry=s3_destination_entry)

add_source_entry(source_type=SFTP_CONNECTOR_TYPE, entry=sftp_source_entry)
add_destination_entry(destination_type=SFTP_CONNECTOR_TYPE, entry=sftp_destination_entry)



# https://developers.notion.com/reference/property-object#url
from dataclasses import dataclass, field
from typing import Optional

from htmlBuilder.tags import Div, HtmlTag, Span

from unstructured_ingest.connector.notion.interfaces import (
    DBCellBase,
    DBPropertyBase,
    FromJSONMixin,
    GetHTMLMixin,
)
from unstructured_ingest.connector.notion.types.date import Date
from unstructured_ingest.connector.notion.types.user import People


@dataclass
class Verification(DBPropertyBase):
    id: str
    name: str
    type: str = "verification"
    verification: dict = field(default_factory=dict)

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)


@dataclass
class VerificationData(FromJSONMixin, GetHTMLMixin):
    state: Optional[str]
    verified_by: Optional[People]
    date: Optional[Date]

    @classmethod
    def from_dict(cls, data: dict):
        verified_by = data.pop("verified_by", None)
        date = data.pop("date", None)
        return cls(
            verified_by=People.from_dict(data=verified_by) if verified_by else None,
            date=Date.from_dict(data=date) if date else None,
            **data,
        )

    def get_html(self) -> Optional[HtmlTag]:
        elements = []
        if state := self.state:
            elements.append(Span([], state))
        if (verified_by := self.verified_by) and (verified_by_html := verified_by.get_html()):
            elements.append(verified_by_html)
        if (date := self.date) and (date_html := date.get_html()):
            elements.append(date_html)
        if elements:
            return Div([], elements)
        return None


@dataclass
class VerificationCell(DBCellBase):
    id: str
    verification: Optional[VerificationData]
    name: Optional[str] = None
    type: str = "verification"

    @classmethod
    def from_dict(cls, data: dict):
        return cls(verification=VerificationData.from_dict(data.pop("verification")), **data)

    def get_html(self) -> Optional[HtmlTag]:
        elements = []
        if name := self.name:
            elements.append(Span([], name))
        if (verification := self.verification) and (verification_html := verification.get_html()):
            elements.append(verification_html)

        if elements:
            return Div([], elements)
        return None



import os
import uuid

import pytest
from googleapiclient.errors import HttpError

from test.integration.connectors.utils.constants import (
    SOURCE_TAG,
    UNCATEGORIZED_TAG,
)
from test.integration.connectors.utils.validation.source import (
    SourceValidationConfigs,
    get_all_file_data,
    run_all_validations,
    update_fixtures,
)
from test.integration.utils import requires_env
from unstructured_ingest.error import (
    SourceConnectionError,
)
from unstructured_ingest.v2.interfaces import Downloader, Indexer
from unstructured_ingest.v2.processes.connectors.google_drive import (
    CONNECTOR_TYPE,
    GoogleDriveAccessConfig,
    GoogleDriveConnectionConfig,
    GoogleDriveDownloader,
    GoogleDriveDownloaderConfig,
    GoogleDriveIndexer,
    GoogleDriveIndexerConfig,
)


@pytest.fixture
def google_drive_connection_config():
    """
    Build a valid GoogleDriveConnectionConfig using the environment variables.
    Expects:
      - GOOGLE_DRIVE_ID
      - GOOGLE_DRIVE_SERVICE_KEY
    """
    drive_id = os.getenv("GOOGLE_DRIVE_ID")
    service_key = os.getenv("GOOGLE_DRIVE_SERVICE_KEY")
    if not drive_id or not service_key:
        pytest.skip("Google Drive credentials not provided in environment variables.")

    access_config = GoogleDriveAccessConfig(service_account_key=service_key)
    return GoogleDriveConnectionConfig(drive_id=drive_id, access_config=access_config)


@pytest.fixture
def google_drive_empty_folder(google_drive_connection_config):
    """
    Creates an empty folder on Google Drive for testing the "empty folder" case.
    The folder is deleted after the test.
    """
    from google.oauth2 import service_account
    from googleapiclient.discovery import build

    access_config = google_drive_connection_config.access_config.get_secret_value()
    creds = service_account.Credentials.from_service_account_info(access_config.service_account_key)
    service = build("drive", "v3", credentials=creds)

    # Create an empty folder.
    file_metadata = {
        "name": f"utic-empty-folder-{uuid.uuid4()}",
        "mimeType": "application/vnd.google-apps.folder",
    }
    folder = service.files().create(body=file_metadata, fields="id, name").execute()
    folder_id = folder.get("id")
    try:
        yield folder_id
    finally:
        service.files().delete(fileId=folder_id).execute()


@requires_env("GOOGLE_DRIVE_SERVICE_KEY")
@pytest.mark.tags(SOURCE_TAG, CONNECTOR_TYPE)
def test_google_drive_source(temp_dir):
    # Retrieve environment variables
    service_account_key = os.environ["GOOGLE_DRIVE_SERVICE_KEY"]

    # Create connection and indexer configurations
    access_config = GoogleDriveAccessConfig(service_account_key=service_account_key)
    connection_config = GoogleDriveConnectionConfig(
        drive_id="1XidSOO76VpZ4m0i3gJN2m1X0Obol3UAi",
        access_config=access_config,
    )
    index_config = GoogleDriveIndexerConfig(recursive=True)

    download_config = GoogleDriveDownloaderConfig(download_dir=temp_dir)

    # Instantiate indexer and downloader
    indexer = GoogleDriveIndexer(
        connection_config=connection_config,
        index_config=index_config,
    )
    downloader = GoogleDriveDownloader(
        connection_config=connection_config,
        download_config=download_config,
    )

    # Run the source connector validation
    source_connector_validation(
        indexer=indexer,
        downloader=downloader,
        configs=SourceValidationConfigs(
            test_id="google_drive_source",
            expected_num_files=1,
            validate_downloaded_files=True,
        ),
    )


@pytest.mark.tags(SOURCE_TAG, CONNECTOR_TYPE, UNCATEGORIZED_TAG)
def source_connector_validation(
    indexer: Indexer,
    downloader: Downloader,
    configs: SourceValidationConfigs,
    overwrite_fixtures: bool = os.getenv("OVERWRITE_FIXTURES", "False").lower() == "true",
) -> None:
    # Run common validations on the process of running a source connector, supporting dynamic
    # validators that get passed in along with comparisons on the saved expected values.
    # If overwrite_fixtures is st to True, will ignore all validators but instead overwrite the
    # expected values with what gets generated by this test.
    all_predownload_file_data = []
    all_postdownload_file_data = []
    indexer.precheck()
    download_dir = downloader.download_config.download_dir
    test_output_dir = configs.test_output_dir()

    for file_data in indexer.run():
        assert file_data
        predownload_file_data = file_data.model_copy(deep=True)
        all_predownload_file_data.append(predownload_file_data)
        resp = downloader.run(file_data=file_data)
        if isinstance(resp, list):
            for r in resp:
                postdownload_file_data = r["file_data"].model_copy(deep=True)
                all_postdownload_file_data.append(postdownload_file_data)
        else:
            postdownload_file_data = resp["file_data"].model_copy(deep=True)
            all_postdownload_file_data.append(postdownload_file_data)

    if not overwrite_fixtures:
        print("Running validation")
        run_all_validations(
            configs=configs,
            predownload_file_data=all_predownload_file_data,
            postdownload_file_data=all_postdownload_file_data,
            download_dir=download_dir,
            test_output_dir=test_output_dir,
        )
    else:
        print("Running fixtures update")
        update_fixtures(
            output_dir=test_output_dir,
            download_dir=download_dir,
            all_file_data=get_all_file_data(
                all_predownload_file_data=all_predownload_file_data,
                all_postdownload_file_data=all_postdownload_file_data,
            ),
            save_downloads=configs.validate_downloaded_files,
            save_filedata=configs.validate_file_data,
        )


# Precheck fails when the drive ID has an appended parameter (simulate copy-paste error)
@pytest.mark.tags("google-drive", "precheck")
@requires_env("GOOGLE_DRIVE_ID", "GOOGLE_DRIVE_SERVICE_KEY")
def test_google_drive_precheck_invalid_parameter(google_drive_connection_config):
    # Append a query parameter as often happens when copying from a URL.
    invalid_drive_id = google_drive_connection_config.drive_id + "?usp=sharing"
    connection_config = GoogleDriveConnectionConfig(
        drive_id=invalid_drive_id,
        access_config=google_drive_connection_config.access_config,
    )
    index_config = GoogleDriveIndexerConfig(recursive=True)
    indexer = GoogleDriveIndexer(connection_config=connection_config, index_config=index_config)
    with pytest.raises(SourceConnectionError) as excinfo:
        indexer.precheck()
    assert "invalid" in str(excinfo.value).lower() or "not found" in str(excinfo.value).lower()


# Precheck fails due to lack of permission (simulate via monkeypatching).
@pytest.mark.tags("google-drive", "precheck")
@requires_env("GOOGLE_DRIVE_ID", "GOOGLE_DRIVE_SERVICE_KEY")
def test_google_drive_precheck_no_permission(google_drive_connection_config, monkeypatch):
    index_config = GoogleDriveIndexerConfig(recursive=True)
    indexer = GoogleDriveIndexer(
        connection_config=google_drive_connection_config,
        index_config=index_config,
    )

    # Monkeypatch get_root_info to always raise an HTTP 403 error.
    def fake_get_root_info(files_client, object_id):
        raise HttpError(
            resp=type("Response", (), {"status": 403, "reason": "Forbidden"})(),
            content=b"Forbidden",
        )

    monkeypatch.setattr(indexer, "get_root_info", fake_get_root_info)
    with pytest.raises(SourceConnectionError) as excinfo:
        indexer.precheck()
    assert "forbidden" in str(excinfo.value).lower() or "permission" in str(excinfo.value).lower()


# Precheck fails when the folder is empty.
# @pytest.mark.tags("google-drive", "precheck")
# @requires_env("GOOGLE_DRIVE_ID", "GOOGLE_DRIVE_SERVICE_KEY")
# def test_google_drive_precheck_empty_folder(
#     google_drive_connection_config, google_drive_empty_folder
# ):
#     # Use the empty folder's ID as the target.
#     connection_config = GoogleDriveConnectionConfig(
#         drive_id=google_drive_empty_folder,
#         access_config=google_drive_connection_config.access_config,
#     )

#     index_config = GoogleDriveIndexerConfig(recursive=True)
#     indexer = GoogleDriveIndexer(connection_config=connection_config, index_config=index_config)
#     with pytest.raises(SourceConnectionError) as excinfo:
#         indexer.precheck()
#     assert "empty folder" in str(excinfo.value).lower()


@pytest.mark.tags("google-drive", "count", "integration")
@requires_env("GOOGLE_DRIVE_ID", "GOOGLE_DRIVE_SERVICE_KEY")
def test_google_drive_count_files(google_drive_connection_config):
    """
    This test verifies that the count_files_recursively method returns the expected count of files.
    According to the test credentials, there are 3 files in the root directory and 1 nested file,
    so the total count should be 4.
    """
    # I assumed that we're applying the same extension filter as with other tests
    # However there's 6 files in total in the test dir
    extensions_filter = ["pdf", "docx"]
    with google_drive_connection_config.get_client() as client:
        count = GoogleDriveIndexer.count_files_recursively(
            client, google_drive_connection_config.drive_id, extensions_filter
        )
    assert count == 4, f"Expected file count of 4, but got {count}"


# Precheck fails with a completely invalid drive ID.
@pytest.mark.tags("google-drive", "precheck")
@requires_env("GOOGLE_DRIVE_ID", "GOOGLE_DRIVE_SERVICE_KEY")
def test_google_drive_precheck_invalid_drive_id(google_drive_connection_config):
    invalid_drive_id = "invalid_drive_id"
    connection_config = GoogleDriveConnectionConfig(
        drive_id=invalid_drive_id,
        access_config=google_drive_connection_config.access_config,
    )
    index_config = GoogleDriveIndexerConfig(recursive=True)
    indexer = GoogleDriveIndexer(connection_config=connection_config, index_config=index_config)
    with pytest.raises(SourceConnectionError) as excinfo:
        indexer.precheck()
    assert "invalid" in str(excinfo.value).lower() or "not found" in str(excinfo.value).lower()



from abc import ABC
from dataclasses import dataclass
from typing import Any, Optional

from pydantic import BaseModel, Field

from unstructured_ingest.utils.data_prep import batch_generator
from unstructured_ingest.utils.dep_check import requires_dependencies

EMBEDDINGS_KEY = "embeddings"


class EmbeddingConfig(BaseModel):
    batch_size: Optional[int] = Field(
        default=32, description="Optional batch size for embedding requests."
    )


@dataclass
class BaseEncoder(ABC):
    config: EmbeddingConfig

    def initialize(self):
        """Initializes the embedding encoder class. Should also validate the instance
        is properly configured: e.g., embed a single a element"""

    def wrap_error(self, e: Exception) -> Exception:
        """Handle errors from the embedding service. Should raise a more informative error
        if possible"""
        return e


@dataclass
class BaseEmbeddingEncoder(BaseEncoder, ABC):
    def initialize(self):
        """Initializes the embedding encoder class. Should also validate the instance
        is properly configured: e.g., embed a single a element"""

    @property
    def dimension(self):
        exemplary_embedding = self.get_exemplary_embedding()
        return len(exemplary_embedding)

    def get_exemplary_embedding(self) -> list[float]:
        return self.embed_query(query="Q")

    @property
    @requires_dependencies(["numpy"])
    def is_unit_vector(self) -> bool:
        """Denotes if the embedding vector is a unit vector."""
        import numpy as np

        exemplary_embedding = self.get_exemplary_embedding()
        return np.isclose(np.linalg.norm(exemplary_embedding), 1.0, rtol=1e-03)

    def get_client(self):
        raise NotImplementedError

    def embed_batch(self, client: Any, batch: list[str]) -> list[list[float]]:
        raise NotImplementedError

    def embed_documents(self, elements: list[dict]) -> list[dict]:
        client = self.get_client()
        elements = elements.copy()
        elements_with_text = [e for e in elements if e.get("text")]
        texts = [e["text"] for e in elements_with_text]
        embeddings = []
        try:
            for batch in batch_generator(texts, batch_size=self.config.batch_size or len(texts)):
                embeddings = self.embed_batch(client=client, batch=batch)
                embeddings.extend(embeddings)
        except Exception as e:
            raise self.wrap_error(e=e)
        for element, embedding in zip(elements_with_text, embeddings):
            element[EMBEDDINGS_KEY] = embedding
        return elements

    def _embed_query(self, query: str) -> list[float]:
        client = self.get_client()
        return self.embed_batch(client=client, batch=[query])[0]

    def embed_query(self, query: str) -> list[float]:
        try:
            return self._embed_query(query=query)
        except Exception as e:
            raise self.wrap_error(e=e)


@dataclass
class AsyncBaseEmbeddingEncoder(BaseEncoder, ABC):
    async def initialize(self):
        """Initializes the embedding encoder class. Should also validate the instance
        is properly configured: e.g., embed a single a element"""

    @property
    async def dimension(self):
        exemplary_embedding = await self.get_exemplary_embedding()
        return len(exemplary_embedding)

    async def get_exemplary_embedding(self) -> list[float]:
        return await self.embed_query(query="Q")

    @property
    @requires_dependencies(["numpy"])
    async def is_unit_vector(self) -> bool:
        """Denotes if the embedding vector is a unit vector."""
        import numpy as np

        exemplary_embedding = await self.get_exemplary_embedding()
        return np.isclose(np.linalg.norm(exemplary_embedding), 1.0, rtol=1e-03)

    def get_client(self):
        raise NotImplementedError

    async def embed_batch(self, client: Any, batch: list[str]) -> list[list[float]]:
        raise NotImplementedError

    async def embed_documents(self, elements: list[dict]) -> list[dict]:
        client = self.get_client()
        elements = elements.copy()
        elements_with_text = [e for e in elements if e.get("text")]
        texts = [e["text"] for e in elements_with_text]
        embeddings = []
        try:
            for batch in batch_generator(texts, batch_size=self.config.batch_size or len(texts)):
                embeddings = await self.embed_batch(client=client, batch=batch)
                embeddings.extend(embeddings)
        except Exception as e:
            raise self.wrap_error(e=e)
        for element, embedding in zip(elements_with_text, embeddings):
            element[EMBEDDINGS_KEY] = embedding
        return elements

    async def _embed_query(self, query: str) -> list[float]:
        client = self.get_client()
        embeddings = await self.embed_batch(client=client, batch=[query])
        return embeddings[0]

    async def embed_query(self, query: str) -> list[float]:
        try:
            return await self._embed_query(query=query)
        except Exception as e:
            raise self.wrap_error(e=e)



import asyncio
import hashlib
from dataclasses import dataclass
from pathlib import Path
from typing import Callable, Optional, TypedDict

from unstructured_ingest.v2.interfaces import UploadStager
from unstructured_ingest.v2.logger import logger
from unstructured_ingest.v2.pipeline.interfaces import PipelineStep
from unstructured_ingest.v2.types.file_data import file_data_from_file
from unstructured_ingest.v2.utils import serialize_base_model_json

STEP_ID = "upload_stage"


class UploadStageStepResponse(TypedDict):
    file_data_path: str
    path: str


@dataclass
class UploadStageStep(PipelineStep):
    process: UploadStager
    identifier: str = STEP_ID

    def __str__(self):
        return f"{self.identifier} ({self.process.__class__.__name__})"

    def __post_init__(self):
        config = (
            self.process.upload_stager_config.model_dump_json()
            if self.process.upload_stager_config
            else None
        )
        self.cache_dir.mkdir(parents=True, exist_ok=True)
        logger.info(f"created {self.identifier} with configs: {config}")

    async def _run_async(
        self, fn: Callable, path: str, file_data_path: str
    ) -> UploadStageStepResponse:
        path = Path(path)
        # Maintain extension
        output_filename = f"{self.get_hash(extras=[path.name])}{path.suffix}"
        fn_kwargs = {
            "elements_filepath": path,
            "file_data": file_data_from_file(path=file_data_path),
            "output_dir": self.cache_dir,
            "output_filename": output_filename,
        }
        if not asyncio.iscoroutinefunction(fn):
            staged_output_path = fn(**fn_kwargs)
        elif semaphore := self.context.semaphore:
            async with semaphore:
                staged_output_path = await fn(**fn_kwargs)
        else:
            staged_output_path = await fn(**fn_kwargs)
        return UploadStageStepResponse(file_data_path=file_data_path, path=str(staged_output_path))

    def get_hash(self, extras: Optional[list[str]]) -> str:
        hashable_string = serialize_base_model_json(
            model=self.process.upload_stager_config, sort_keys=True, ensure_ascii=True
        )
        if extras:
            hashable_string += "".join(extras)
        return hashlib.sha256(hashable_string.encode()).hexdigest()[:12]



import typing as t

from .astradb import AstraDBWriter
from .azure_ai_search import AzureAiSearchWriter
from .base_writer import Writer
from .chroma import ChromaWriter
from .clarifai import ClarifaiWriter
from .databricks_volumes import DatabricksVolumesWriter
from .delta_table import DeltaTableWriter
from .elasticsearch import ElasticsearchWriter
from .fsspec.azure import AzureWriter
from .fsspec.box import BoxWriter
from .fsspec.dropbox import DropboxWriter
from .fsspec.gcs import GcsWriter
from .fsspec.s3 import S3Writer
from .kafka import KafkaWriter
from .mongodb import MongodbWriter
from .opensearch import OpenSearchWriter
from .pinecone import PineconeWriter
from .qdrant import QdrantWriter
from .sql import SqlWriter
from .vectara import VectaraWriter
from .weaviate import WeaviateWriter

writer_map: t.Dict[str, t.Type[Writer]] = {
    "astradb": AstraDBWriter,
    "azure": AzureWriter,
    "azure_ai_search": AzureAiSearchWriter,
    "box": BoxWriter,
    "chroma": ChromaWriter,
    "clarifai": ClarifaiWriter,
    "databricks_volumes": DatabricksVolumesWriter,
    "delta_table": DeltaTableWriter,
    "dropbox": DropboxWriter,
    "elasticsearch": ElasticsearchWriter,
    "gcs": GcsWriter,
    "kafka": KafkaWriter,
    "mongodb": MongodbWriter,
    "opensearch": OpenSearchWriter,
    "pinecone": PineconeWriter,
    "qdrant": QdrantWriter,
    "s3": S3Writer,
    "sql": SqlWriter,
    "vectara": VectaraWriter,
    "weaviate": WeaviateWriter,
}

__all__ = ["writer_map"]



import typing as t
from dataclasses import fields, is_dataclass
from gettext import gettext as _

import click

from unstructured_ingest.cli.interfaces import (
    CliChunkingConfig,
    CliConfig,
    CliEmbeddingConfig,
    CliPartitionConfig,
    CliPermissionsConfig,
    CliProcessorConfig,
    CliReadConfig,
    CliRetryStrategyConfig,
)
from unstructured_ingest.interfaces import BaseConfig
from unstructured_ingest.logger import logger


def conform_click_options(options: dict):
    # Click sets all multiple fields as tuple, this needs to be updated to list
    for k, v in options.items():
        if isinstance(v, tuple):
            options[k] = list(v)


def extract_config(flat_data: dict, config: t.Type[BaseConfig]) -> BaseConfig:
    """
    To be able to extract a nested dataclass from a flat dictionary (as in one coming
    from a click-based options input), the config class is dynamically looked through for
    nested dataclass fields and new nested dictionaries are created to conform to the
    shape the overall class expects when parsing from a dict. During the process, this will create
    copies of the original dictionary to avoid pruning fields but this isn't a
    problem since the `from_dict()` method ignores unneeded values.

    Not handling more complex edge cases for now such as nested types i.e Union[List[List[...]]]
    """

    def conform_dict(inner_d: dict, inner_config: t.Type[BaseConfig]):
        # Catch edge cases (i.e. Dict[str, ...]) where underlying type is not a concrete Class,
        # causing 'issubclass() arg 1 must be a class' errors, return False
        def is_subclass(instance, class_type) -> bool:
            try:
                return issubclass(instance, class_type)
            except Exception:
                return False

        dd = inner_d.copy()
        for field in fields(inner_config):
            f_type = field.type
            # Handle the case where the type of a value if a Union (possibly optional)
            if t.get_origin(f_type) is t.Union:
                union_values = t.get_args(f_type)
                # handle List types
                union_values = [
                    t.get_args(u)[0] if t.get_origin(u) is list else u for u in union_values
                ]
                # Ignore injected NoneType when optional
                concrete_union_values = [v for v in union_values if not is_subclass(v, type(None))]
                dataclass_union_values = [v for v in concrete_union_values if is_dataclass(v)]
                non_dataclass_union_values = [
                    v for v in concrete_union_values if not is_dataclass(v)
                ]
                if not dataclass_union_values:
                    continue
                # Check if the key for this field already exists in the dictionary,
                # if so it might map to one of these non dataclass fields and this
                # can't be enforced
                if non_dataclass_union_values and field.name in dd:
                    continue
                if len(dataclass_union_values) > 1:
                    logger.warning(
                        "more than one dataclass type possible for field {}, "
                        "not extracting: {}".format(field.name, ", ".join(dataclass_union_values))
                    )
                    continue
                f_type = dataclass_union_values[0]
            origin = t.get_origin(f_type)
            if origin:
                f_type = origin
            if is_subclass(f_type, BaseConfig):
                dd[field.name] = conform_dict(inner_d=dd, inner_config=f_type)
        return dd

    adjusted_dict = conform_dict(inner_d=flat_data, inner_config=config)
    return config.from_dict(adjusted_dict, apply_name_overload=False)


def extract_configs(
    data: dict,
    extras: t.Optional[t.Dict[str, t.Type[BaseConfig]]] = None,
    validate: t.Optional[t.List[t.Type[BaseConfig]]] = None,
    add_defaults: bool = True,
) -> t.Dict[str, BaseConfig]:
    """
    Extract all common configs used across CLI command and validate that any
    command-specific configs have all their needed information from the Click
    options that are passed in during invocation.
    """
    validate = validate if validate else []
    res = (
        {
            "read_config": extract_config(flat_data=data, config=CliReadConfig),
            "partition_config": extract_config(flat_data=data, config=CliPartitionConfig),
            "embedding_config": extract_config(flat_data=data, config=CliEmbeddingConfig),
            "chunking_config": extract_config(flat_data=data, config=CliChunkingConfig),
            "processor_config": extract_config(flat_data=data, config=CliProcessorConfig),
            "permissions_config": extract_config(flat_data=data, config=CliPermissionsConfig),
            "retry_strategy_config": extract_config(flat_data=data, config=CliRetryStrategyConfig),
        }
        if add_defaults
        else {}
    )
    if extras:
        for k, conf in extras.items():
            try:
                res[k] = extract_config(flat_data=data, config=conf)
            except Exception as e:
                logger.error(f"failed to extract config from {conf.__name__}")
                raise e
    for v in validate:
        try:
            extract_config(flat_data=data, config=v)
        except Exception as e:
            raise Exception(f"failed to validate config {v.__name__}") from e

    return res


def add_options(
    cmd: click.Command, extras: t.List[t.Type[CliConfig]], is_src: bool = True
) -> click.Command:
    configs: t.List[t.Type[CliConfig]] = (
        [
            CliPartitionConfig,
            CliReadConfig,
            CliEmbeddingConfig,
            CliChunkingConfig,
            CliProcessorConfig,
            CliPermissionsConfig,
            CliRetryStrategyConfig,
        ]
        if is_src
        else []
    )
    # make sure what's unique to this cmd appears first
    extras.extend(configs)
    for config in extras:
        try:
            config.add_cli_options(cmd=cmd)
        except ValueError as e:
            raise ValueError(f"failed to set configs from {config.__name__}: {e}")
    return cmd


class Group(click.Group):
    def parse_args(self, ctx, args):
        """
        This allows for subcommands to be called with the --help flag without breaking
        if parent command is missing any of its required parameters
        """

        try:
            return super().parse_args(ctx, args)
        except click.MissingParameter:
            if "--help" not in args:
                raise

            # remove the required params so that help can display
            for param in self.params:
                param.required = False
            return super().parse_args(ctx, args)

    def format_commands(self, ctx: click.Context, formatter: click.HelpFormatter) -> None:
        """
        Copy of the original click.Group format_commands() method but replacing
        'Commands' -> 'Destinations'
        """
        commands = []
        for subcommand in self.list_commands(ctx):
            cmd = self.get_command(ctx, subcommand)
            # What is this, the tool lied about a command.  Ignore it
            if cmd is None:
                continue
            if cmd.hidden:
                continue

            commands.append((subcommand, cmd))

        # allow for 3 times the default spacing
        if len(commands):
            if formatter.width:
                limit = formatter.width - 6 - max(len(cmd[0]) for cmd in commands)
            else:
                limit = -6 - max(len(cmd[0]) for cmd in commands)

            rows = []
            for subcommand, cmd in commands:
                help = cmd.get_short_help_str(limit)
                rows.append((subcommand, help))

            if rows:
                with formatter.section(_("Destinations")):
                    formatter.write_dl(rows)



from abc import ABC
from functools import wraps


class CustomError(Exception, ABC):
    error_string: str

    @classmethod
    def wrap(cls, f):
        """
        Provides a wrapper for a function that catches any exception and
        re-raises it as the customer error. If the exception itself is already an instance
        of the custom error, re-raises original error.
        """

        @wraps(f)
        def wrapper(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except BaseException as error:
                if not isinstance(error, cls) and not issubclass(type(error), cls):
                    raise cls(cls.error_string.format(str(error))) from error
                raise

        return wrapper


class SourceConnectionError(CustomError):
    error_string = "Error in getting data from upstream data source: {}"


class SourceConnectionNetworkError(SourceConnectionError):
    error_string = "Error in connecting to upstream data source: {}"


class DestinationConnectionError(CustomError):
    error_string = "Error in connecting to downstream data source: {}"


class EmbeddingEncoderConnectionError(CustomError):
    error_string = "Error in connecting to the embedding model provider: {}"


class WriteError(CustomError):
    error_string = "Error in writing to downstream data source: {}"


class PartitionError(CustomError):
    error_string = "Error in partitioning content: {}"



import logging
import typing as t
from dataclasses import dataclass

import click

from unstructured_ingest.cli.base.cmd import BaseCmd
from unstructured_ingest.cli.cmd_factory import get_src_cmd
from unstructured_ingest.cli.common import (
    log_options,
)
from unstructured_ingest.cli.interfaces import BaseConfig, CliFilesStorageConfig
from unstructured_ingest.cli.utils import (
    add_options,
    conform_click_options,
    extract_config,
    extract_configs,
)
from unstructured_ingest.logger import ingest_log_streaming_init, logger
from unstructured_ingest.runner.writers import writer_map


@dataclass
class BaseDestCmd(BaseCmd):
    write_config: t.Optional[t.Type[BaseConfig]] = None

    def get_dest_runner(self, source_cmd: str, options: dict, parent_options: dict):
        src_cmd_fn = get_src_cmd(cmd_name=source_cmd)
        src_cmd = src_cmd_fn()
        runner = src_cmd.get_source_runner(options=parent_options)
        addition_configs = self.addition_configs
        if "connector_config" not in addition_configs:
            addition_configs["connector_config"] = self.cli_config
        if self.write_config:
            addition_configs["write_config"] = self.write_config
        configs = extract_configs(
            options,
            validate=[self.cli_config] if self.cli_config else None,
            extras=addition_configs,
            add_defaults=False,
        )
        writer_cls = writer_map[self.cmd_name_key]
        writer = writer_cls(**configs)  # type: ignore
        runner.writer = writer
        runner.writer_kwargs = options
        return runner

    def check_dest_options(self, options: dict):
        extract_config(flat_data=options, config=self.cli_config)

    def dest(self, ctx: click.Context, **options):
        if not ctx.parent:
            raise click.ClickException("destination command called without a parent")
        if not ctx.parent.info_name:
            raise click.ClickException("parent command missing info name")
        source_cmd = ctx.parent.info_name.replace("-", "_")
        parent_options: dict = ctx.parent.params if ctx.parent else {}
        conform_click_options(options)
        verbose = parent_options.get("verbose", False)
        ingest_log_streaming_init(logging.DEBUG if verbose else logging.INFO)
        log_options(parent_options, verbose=verbose)
        log_options(options, verbose=verbose)
        try:
            self.check_dest_options(options=options)
            runner = self.get_dest_runner(
                source_cmd=source_cmd,
                options=options,
                parent_options=parent_options,
            )
            runner.run(**parent_options)
        except Exception as e:
            logger.error(e, exc_info=True)
            raise click.ClickException(str(e)) from e

    def get_dest_cmd(self) -> click.Command:
        # Dynamically create the command without the use of click decorators
        fn = self.dest
        fn = click.pass_context(fn)
        cmd: click.Group = click.command(fn)
        cmd.name = self.cmd_name
        cmd.invoke_without_command = True
        options = [self.cli_config] if self.cli_config else []
        options += self.additional_cli_options
        if self.is_fsspec and CliFilesStorageConfig not in options:
            options.append(CliFilesStorageConfig)
        add_options(cmd, extras=options, is_src=False)
        return cmd



from __future__ import annotations

import asyncio
import json
from abc import ABC, abstractmethod
from contextlib import asynccontextmanager
from dataclasses import dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING, Any, AsyncGenerator, Optional

from pydantic import Field

from unstructured_ingest.error import DestinationConnectionError
from unstructured_ingest.logger import logger
from unstructured_ingest.utils.data_prep import flatten_dict
from unstructured_ingest.utils.dep_check import requires_dependencies
from unstructured_ingest.v2.constants import RECORD_ID_LABEL
from unstructured_ingest.v2.interfaces import (
    ConnectionConfig,
    Uploader,
    UploaderConfig,
    UploadStager,
    UploadStagerConfig,
)
from unstructured_ingest.v2.types.file_data import FileData

CONNECTOR_TYPE = "lancedb"

if TYPE_CHECKING:
    from lancedb import AsyncConnection
    from lancedb.table import AsyncTable
    from pandas import DataFrame


class LanceDBConnectionConfig(ConnectionConfig, ABC):
    uri: str = Field(description="The uri of the database.")

    @abstractmethod
    def get_storage_options(self) -> Optional[dict[str, str]]:
        raise NotImplementedError

    @asynccontextmanager
    @requires_dependencies(["lancedb"], extras="lancedb")
    @DestinationConnectionError.wrap
    async def get_async_connection(self) -> AsyncGenerator["AsyncConnection", None]:
        import lancedb

        with await lancedb.connect_async(
            self.uri,
            storage_options=self.get_storage_options(),
        ) as connection:
            yield connection


class LanceDBRemoteConnectionConfig(LanceDBConnectionConfig):
    timeout: str = Field(
        default="30s",
        description=(
            "Timeout for the entire request, from connection until the response body has finished"
            "in a [0-9]+(ns|us|ms|[smhdwy]) format."
        ),
        pattern=r"[0-9]+(ns|us|ms|[smhdwy])",
    )


class LanceDBUploadStagerConfig(UploadStagerConfig):
    pass


@dataclass
class LanceDBUploadStager(UploadStager):
    upload_stager_config: LanceDBUploadStagerConfig = field(
        default_factory=LanceDBUploadStagerConfig
    )

    @requires_dependencies(["pandas"], extras="lancedb")
    def run(
        self,
        elements_filepath: Path,
        file_data: FileData,
        output_dir: Path,
        output_filename: str,
        **kwargs: Any,
    ) -> Path:
        import pandas as pd

        with open(elements_filepath) as elements_file:
            elements_contents: list[dict] = json.load(elements_file)

        df = pd.DataFrame(
            [
                self.conform_dict(element_dict=element_dict, file_data=file_data)
                for element_dict in elements_contents
            ]
        )

        output_path = (output_dir / output_filename).with_suffix(".feather")
        df.to_feather(output_path)

        return output_path

    def conform_dict(self, element_dict: dict, file_data: FileData) -> dict:
        data = element_dict.copy()
        return {
            "vector": data.pop("embeddings", None),
            RECORD_ID_LABEL: file_data.identifier,
            **flatten_dict(data, separator="-"),
        }


class LanceDBUploaderConfig(UploaderConfig):
    table_name: str = Field(description="The name of the table.")


@dataclass
class LanceDBUploader(Uploader):
    upload_config: LanceDBUploaderConfig
    connection_config: LanceDBConnectionConfig
    connector_type: str = CONNECTOR_TYPE

    @DestinationConnectionError.wrap
    def precheck(self):
        async def _precheck() -> None:
            async with self.connection_config.get_async_connection() as conn:
                table = await conn.open_table(self.upload_config.table_name)
                table.close()

        asyncio.run(_precheck())

    @asynccontextmanager
    async def get_table(self) -> AsyncGenerator["AsyncTable", None]:
        async with self.connection_config.get_async_connection() as conn:
            table = await conn.open_table(self.upload_config.table_name)
            try:
                yield table
            finally:
                table.close()

    @requires_dependencies(["pandas"], extras="lancedb")
    async def run_async(self, path, file_data, **kwargs):
        import pandas as pd

        df = pd.read_feather(path)
        async with self.get_table() as table:
            schema = await table.schema()
            df = self._fit_to_schema(df, schema)
            if RECORD_ID_LABEL not in schema.names:
                logger.warning(
                    f"Designated table doesn't contain {RECORD_ID_LABEL} column of type"
                    " string which is required to support overwriting updates on subsequent"
                    " uploads of the same record. New rows will be appended instead."
                )
            else:
                await table.delete(f'{RECORD_ID_LABEL} = "{file_data.identifier}"')
            await table.add(data=df)

    def _fit_to_schema(self, df: "DataFrame", schema) -> "DataFrame":
        import pandas as pd

        columns = set(df.columns)
        schema_fields = set(schema.names)
        columns_to_drop = columns - schema_fields
        missing_columns = schema_fields - columns

        if columns_to_drop:
            logger.info(
                "Following columns will be dropped to match the table's schema: "
                f"{', '.join(columns_to_drop)}"
            )
        if missing_columns:
            logger.info(
                "Following null filled columns will be added to match the table's schema:"
                f" {', '.join(missing_columns)} "
            )

        df = df.drop(columns=columns_to_drop)

        for column in missing_columns:
            df[column] = pd.Series()

        return df



from pathlib import Path

import pandas as pd
import pytest
from pytest_mock import MockerFixture

from unstructured_ingest.v2.processes.connectors.sql.sql import (
    SQLConnectionConfig,
    SQLUploader,
    SQLUploaderConfig,
    SQLUploadStager,
)
from unstructured_ingest.v2.types.file_data import FileData, SourceIdentifiers


@pytest.fixture
def mock_instance() -> SQLUploadStager:
    return SQLUploadStager()


@pytest.fixture
def mock_uploader(mocker: MockerFixture) -> SQLUploader:
    mock_connection_config = mocker.Mock(spec=SQLConnectionConfig)
    mock_upload_config = mocker.Mock(spec=SQLUploaderConfig)
    return SQLUploader(
        upload_config=mock_upload_config,
        connection_config=mock_connection_config,
        connector_type="sql_test",
    )


@pytest.mark.parametrize(
    ("input_filepath", "output_filename", "expected"),
    [
        (
            "/path/to/input_file.ndjson",
            "output_file.ndjson",
            "output_file.ndjson",
        ),
        ("input_file.txt", "output_file.json", "output_file.txt"),
        ("/path/to/input_file.json", "output_file", "output_file.json"),
    ],
)
def test_run_output_filename_suffix(
    mocker: MockerFixture,
    mock_instance: SQLUploadStager,
    input_filepath: str,
    output_filename: str,
    expected: str,
):
    output_dir = Path("/tmp/test/output_dir")

    # Mocks
    mock_get_data = mocker.patch(
        "unstructured_ingest.v2.processes.connectors.sql.sql.get_data",
        return_value=[{"key": "value"}, {"key": "value2"}],
    )
    mock_conform_dict = mocker.patch.object(
        SQLUploadStager, "conform_dict", side_effect=lambda element_dict, file_data: element_dict
    )
    mock_conform_dataframe = mocker.patch.object(
        SQLUploadStager, "conform_dataframe", side_effect=lambda df: df
    )
    mock_get_output_path = mocker.patch.object(
        SQLUploadStager, "get_output_path", return_value=output_dir / expected
    )
    mock_write_output = mocker.patch(
        "unstructured_ingest.v2.processes.connectors.sql.sql.write_data", return_value=None
    )

    # Act
    result = mock_instance.run(
        elements_filepath=Path(input_filepath),
        file_data=FileData(
            identifier="test",
            connector_type="test",
            source_identifiers=SourceIdentifiers(filename=input_filepath, fullpath=input_filepath),
        ),
        output_dir=output_dir,
        output_filename=output_filename,
    )

    # Assert
    mock_get_data.assert_called_once_with(path=Path(input_filepath))
    assert mock_conform_dict.call_count == 2
    mock_conform_dataframe.assert_called_once()
    mock_get_output_path.assert_called_once_with(output_filename=expected, output_dir=output_dir)
    mock_write_output.assert_called_once_with(
        path=output_dir / expected, data=[{"key": "value"}, {"key": "value2"}]
    )
    assert result.name == expected


def test_fit_to_schema_drop_columns(mocker: MockerFixture, mock_uploader: SQLUploader):
    df = pd.DataFrame(
        {
            "col1": [1, 2],
            "col2": [3, 4],
            "col3": [5, 6],
        }
    )
    mocker.patch.object(mock_uploader, "get_table_columns", return_value=["col1", "col2"])

    result = mock_uploader._fit_to_schema(df)

    assert "col3" not in result.columns
    assert "col1" in result.columns
    assert "col2" in result.columns


def test_fit_to_schema_add_missing_columns(mocker: MockerFixture, mock_uploader: SQLUploader):
    df = pd.DataFrame(
        {
            "col1": [1, 2],
        }
    )
    mocker.patch.object(mock_uploader, "get_table_columns", return_value=["col1", "col2"])

    result = mock_uploader._fit_to_schema(df)

    assert "col2" in result.columns
    assert result["col2"].isnull().all()


def test_fit_to_schema_no_changes(mocker: MockerFixture, mock_uploader: SQLUploader):
    df = pd.DataFrame(
        {
            "col1": [1, 2],
            "col2": [3, 4],
        }
    )
    mocker.patch.object(mock_uploader, "get_table_columns", return_value=["col1", "col2"])

    result = mock_uploader._fit_to_schema(df)

    assert "col1" in result.columns
    assert "col2" in result.columns
    assert result.equals(df)


def test_fit_to_schema_no_add_missing_columns(mocker: MockerFixture, mock_uploader: SQLUploader):
    df = pd.DataFrame(
        {
            "col1": [1, 2],
        }
    )
    mocker.patch.object(mock_uploader, "get_table_columns", return_value=["col1", "col2"])

    result = mock_uploader._fit_to_schema(df, add_missing_columns=False)

    assert "col2" not in result.columns
    assert "col1" in result.columns



import json
import math
import os
import re
import time
from pathlib import Path
from typing import Generator
from uuid import uuid4

import pytest
from _pytest.fixtures import TopRequest
from pinecone import Pinecone, ServerlessSpec
from pinecone.core.openapi.shared.exceptions import NotFoundException

from test.integration.connectors.utils.constants import DESTINATION_TAG, VECTOR_DB_TAG
from test.integration.connectors.utils.validation.destination import (
    StagerValidationConfigs,
    stager_validation,
)
from test.integration.utils import requires_env
from unstructured_ingest.error import DestinationConnectionError
from unstructured_ingest.v2.logger import logger
from unstructured_ingest.v2.processes.connectors.pinecone import (
    CONNECTOR_TYPE,
    MAX_QUERY_RESULTS,
    PineconeAccessConfig,
    PineconeConnectionConfig,
    PineconeUploader,
    PineconeUploaderConfig,
    PineconeUploadStager,
    PineconeUploadStagerConfig,
)
from unstructured_ingest.v2.types.file_data import FileData, SourceIdentifiers

METADATA_BYTES_LIMIT = (
    40960  # 40KB https://docs.pinecone.io/reference/quotas-and-limits#hard-limits
)
VECTOR_DIMENSION = 384
SPEC = {"serverless": {"cloud": "aws", "region": "us-east-1"}}
ALLOWED_METADATA_FIELD = "text"
API_KEY = "PINECONE_API_KEY"


def get_api_key() -> str:
    api_key = os.getenv(API_KEY, None)
    assert api_key
    return api_key


def wait_for_delete(client: Pinecone, index_name: str, timeout=60, interval=1) -> None:
    start = time.time()
    while True and time.time() - start < timeout:
        try:
            description = client.describe_index(name=index_name)
            logger.info(f"current index status: {description}")
        except NotFoundException:
            return
        time.sleep(interval)

    raise TimeoutError("time out waiting for index to delete")


def wait_for_ready(client: Pinecone, index_name: str, timeout=60, interval=1) -> None:
    def is_ready_status():
        description = client.describe_index(name=index_name)
        status = description["status"]
        return status["ready"]

    start = time.time()
    is_ready = is_ready_status()
    while not is_ready and time.time() - start < timeout:
        time.sleep(interval)
        is_ready = is_ready_status()
    if not is_ready:
        raise TimeoutError("time out waiting for index to be ready")


@pytest.fixture
def pinecone_index() -> Generator[str, None, None]:
    pinecone = Pinecone(api_key=get_api_key())
    random_id = str(uuid4()).split("-")[0]
    index_name = f"ingest-test-{random_id}"
    assert len(index_name) < 45
    logger.info(f"Creating index: {index_name}")
    try:
        pinecone.create_index(
            name=index_name,
            dimension=384,
            metric="cosine",
            spec=ServerlessSpec(
                cloud="aws",
                region="us-east-1",
            ),
            deletion_protection="disabled",
        )
        wait_for_ready(client=pinecone, index_name=index_name)
        yield index_name
    except Exception as e:
        logger.error(f"failed to create index {index_name}: {e}")
    finally:
        try:
            logger.info(f"deleting index: {index_name}")
            pinecone.delete_index(name=index_name)
            wait_for_delete(client=pinecone, index_name=index_name)
        except NotFoundException:
            return


def validate_pinecone_index(
    index_name: str,
    expected_num_of_vectors: int,
    retries=30,
    interval=1,
    namespace: str = "default",
) -> None:
    # Because there's a delay for the index to catch up to the recent writes, add in a retry
    pinecone = Pinecone(api_key=get_api_key())
    index = pinecone.Index(name=index_name, namespace=namespace)
    vector_count = -1
    for i in range(retries):
        index_stats = index.describe_index_stats()
        vector_count = index_stats["total_vector_count"]
        if vector_count == expected_num_of_vectors:
            logger.info(f"expected {expected_num_of_vectors} == vector count {vector_count}")
            break
        logger.info(
            f"retry attempt {i}: expected {expected_num_of_vectors} != vector count {vector_count}"
        )
        time.sleep(interval)
    assert vector_count == expected_num_of_vectors, (
        f"vector count from index ({vector_count}) doesn't "
        f"match expected number: {expected_num_of_vectors}"
    )


@requires_env(API_KEY)
@pytest.mark.asyncio
@pytest.mark.tags(CONNECTOR_TYPE, DESTINATION_TAG, VECTOR_DB_TAG)
async def test_pinecone_destination(pinecone_index: str, upload_file: Path, temp_dir: Path):

    file_data = FileData(
        source_identifiers=SourceIdentifiers(fullpath=upload_file.name, filename=upload_file.name),
        connector_type=CONNECTOR_TYPE,
        identifier="pinecone_mock_id",
    )

    connection_config = PineconeConnectionConfig(
        index_name=pinecone_index,
        access_config=PineconeAccessConfig(api_key=get_api_key()),
    )
    stager_config = PineconeUploadStagerConfig()
    stager = PineconeUploadStager(upload_stager_config=stager_config)
    new_upload_file = stager.run(
        elements_filepath=upload_file,
        output_dir=temp_dir,
        output_filename=upload_file.name,
        file_data=file_data,
    )

    upload_config = PineconeUploaderConfig()
    uploader = PineconeUploader(connection_config=connection_config, upload_config=upload_config)
    uploader.precheck()

    uploader.run(path=new_upload_file, file_data=file_data)
    with new_upload_file.open() as f:
        staged_content = json.load(f)
    expected_num_of_vectors = len(staged_content)
    logger.info("validating first upload")
    validate_pinecone_index(
        index_name=pinecone_index, expected_num_of_vectors=expected_num_of_vectors
    )

    # Rerun uploader and make sure no duplicates exist
    uploader.run(path=new_upload_file, file_data=file_data)
    logger.info("validating second upload")
    validate_pinecone_index(
        index_name=pinecone_index, expected_num_of_vectors=expected_num_of_vectors
    )


@requires_env(API_KEY)
@pytest.mark.asyncio
@pytest.mark.tags(CONNECTOR_TYPE, DESTINATION_TAG, VECTOR_DB_TAG)
@pytest.mark.skip(reason="TODO: get this to work")
async def test_pinecone_destination_large_index(
    pinecone_index: str, upload_file: Path, temp_dir: Path
):
    new_file = temp_dir / "large_file.json"
    with upload_file.open() as f:
        upload_content = json.load(f)

    min_entries = math.ceil((MAX_QUERY_RESULTS * 2) / len(upload_content))
    new_content = (upload_content * min_entries)[: (2 * MAX_QUERY_RESULTS)]
    print(f"Creating large index content with {len(new_content)} records")
    with new_file.open("w") as f:
        json.dump(new_content, f)

    expected_num_of_vectors = len(new_content)
    file_data = FileData(
        source_identifiers=SourceIdentifiers(fullpath=new_file.name, filename=new_file.name),
        connector_type=CONNECTOR_TYPE,
        identifier="pinecone_mock_id",
    )
    connection_config = PineconeConnectionConfig(
        index_name=pinecone_index,
        access_config=PineconeAccessConfig(api_key=get_api_key()),
    )
    stager_config = PineconeUploadStagerConfig()
    stager = PineconeUploadStager(upload_stager_config=stager_config)
    new_upload_file = stager.run(
        elements_filepath=new_file,
        output_dir=temp_dir,
        output_filename=new_file.name,
        file_data=file_data,
    )

    upload_config = PineconeUploaderConfig()
    uploader = PineconeUploader(connection_config=connection_config, upload_config=upload_config)
    uploader.precheck()

    uploader.run(path=new_upload_file, file_data=file_data)
    validate_pinecone_index(
        index_name=pinecone_index, expected_num_of_vectors=expected_num_of_vectors
    )
    # Rerun uploader and make sure no duplicates exist
    uploader.run(path=new_upload_file, file_data=file_data)
    logger.info("validating second upload")
    validate_pinecone_index(
        index_name=pinecone_index, expected_num_of_vectors=expected_num_of_vectors
    )


@requires_env(API_KEY)
@pytest.mark.asyncio
@pytest.mark.tags(CONNECTOR_TYPE, DESTINATION_TAG, VECTOR_DB_TAG)
async def test_pinecone_destination_namespace(
    pinecone_index: str, upload_file: Path, temp_dir: Path
):
    """
    tests namespace functionality of destination connector.
    """

    # creates a file data structure.
    file_data = FileData(
        source_identifiers=SourceIdentifiers(fullpath=upload_file.name, filename=upload_file.name),
        connector_type=CONNECTOR_TYPE,
        identifier="pinecone_mock_id",
    )

    connection_config = PineconeConnectionConfig(
        index_name=pinecone_index,
        access_config=PineconeAccessConfig(api_key=get_api_key()),
    )

    stager_config = PineconeUploadStagerConfig()

    stager = PineconeUploadStager(upload_stager_config=stager_config)
    new_upload_file = stager.run(
        elements_filepath=upload_file,
        output_dir=temp_dir,
        output_filename=upload_file.name,
        file_data=file_data,
    )

    # here add namespace defintion
    upload_config = PineconeUploaderConfig()
    namespace_test_name = "user-1"
    upload_config.namespace = namespace_test_name
    uploader = PineconeUploader(connection_config=connection_config, upload_config=upload_config)
    uploader.precheck()

    uploader.run(path=new_upload_file, file_data=file_data)
    with new_upload_file.open() as f:
        staged_content = json.load(f)
    expected_num_of_vectors = len(staged_content)
    logger.info("validating first upload")
    validate_pinecone_index(
        index_name=pinecone_index,
        expected_num_of_vectors=expected_num_of_vectors,
        namespace=namespace_test_name,
    )

    # Rerun uploader and make sure no duplicates exist
    uploader.run(path=new_upload_file, file_data=file_data)
    logger.info("validating second upload")
    validate_pinecone_index(
        index_name=pinecone_index,
        expected_num_of_vectors=expected_num_of_vectors,
        namespace=namespace_test_name,
    )


@requires_env(API_KEY)
@pytest.mark.tags(CONNECTOR_TYPE, DESTINATION_TAG, VECTOR_DB_TAG)
def test_large_metadata(pinecone_index: str, tmp_path: Path, upload_file: Path):
    stager = PineconeUploadStager()
    uploader = PineconeUploader(
        connection_config=PineconeConnectionConfig(
            access_config=PineconeAccessConfig(api_key=get_api_key()),
            index_name=pinecone_index,
        ),
        upload_config=PineconeUploaderConfig(),
    )
    large_metadata_upload_file = tmp_path / "mock-upload-file.pdf.json"
    large_metadata = {ALLOWED_METADATA_FIELD: "0" * 2 * METADATA_BYTES_LIMIT}

    with open(upload_file) as file:
        elements = json.load(file)

    with open(large_metadata_upload_file, "w") as file:
        mock_element = elements[0]
        mock_element["metadata"] = large_metadata
        json.dump([mock_element], file)

    file_data = FileData(
        source_identifiers=SourceIdentifiers(
            fullpath=large_metadata_upload_file.name, filename=large_metadata_upload_file.name
        ),
        connector_type=CONNECTOR_TYPE,
        identifier="mock-file-data",
    )
    staged_file = stager.run(
        elements_filepath=large_metadata_upload_file,
        file_data=file_data,
        output_dir=tmp_path,
        output_filename=large_metadata_upload_file.name,
    )
    try:
        uploader.run(staged_file, file_data)
    except DestinationConnectionError as e:
        error_line = r"Metadata size is \d+ bytes, which exceeds the limit of \d+ bytes per vector"
        if re.search(re.compile(error_line), str(e)) is None:
            raise e
        raise pytest.fail("Upload request failed due to metadata exceeding limits.")

    validate_pinecone_index(pinecone_index, 1, interval=5)


@pytest.mark.tags(CONNECTOR_TYPE, DESTINATION_TAG, VECTOR_DB_TAG)
@pytest.mark.parametrize("upload_file_str", ["upload_file_ndjson", "upload_file"])
def test_pinecone_stager(
    request: TopRequest,
    upload_file_str: str,
    tmp_path: Path,
):
    upload_file: Path = request.getfixturevalue(upload_file_str)
    stager = PineconeUploadStager()
    stager_validation(
        configs=StagerValidationConfigs(test_id=CONNECTOR_TYPE, expected_count=22),
        input_file=upload_file,
        stager=stager,
        tmp_dir=tmp_path,
    )


@requires_env(API_KEY)
@pytest.mark.tags(CONNECTOR_TYPE, DESTINATION_TAG, VECTOR_DB_TAG)
def test_pinecone_create_destination(pinecone_index):
    uploader = PineconeUploader(
        connection_config=PineconeConnectionConfig(
            access_config=PineconeAccessConfig(api_key=get_api_key())
        ),
        upload_config=PineconeUploaderConfig(),
    )

    random_id = str(uuid4()).split("-")[0]

    index_name = f"test-create-destination-{random_id}"

    assert not uploader.index_exists(index_name=index_name)

    try:
        uploader.create_destination(destination_name=index_name, vector_length=1536)
    except Exception as e:
        error_body = getattr(e, "body", None)
        raise pytest.fail(f"failed to create destination: {e} {error_body}")

    assert uploader.index_exists(index_name=index_name), "destination was not created successfully"

    try:
        pc = uploader.connection_config.get_client()
        logger.info(f"deleting index for test create destination: {index_name}")
        pc.delete_index(name=index_name)
    except Exception as e:
        raise pytest.fail(f"failed to cleanup / delete the destination: {e}")

    assert not uploader.index_exists(index_name=index_name), "cleanup failed"



import os
import uuid
from pathlib import Path

import pytest
from office365.graph_client import GraphClient

from test.integration.connectors.utils.constants import (
    BLOB_STORAGE_TAG,
    DESTINATION_TAG,
    SOURCE_TAG,
)
from test.integration.connectors.utils.validation.source import (
    SourceValidationConfigs,
    source_connector_validation,
)
from test.integration.utils import requires_env
from unstructured_ingest.v2.processes.connectors.onedrive import (
    CONNECTOR_TYPE,
    OnedriveAccessConfig,
    OnedriveConnectionConfig,
    OnedriveDownloader,
    OnedriveDownloaderConfig,
    OnedriveIndexer,
    OnedriveIndexerConfig,
    OnedriveUploader,
    OnedriveUploaderConfig,
)
from unstructured_ingest.v2.types.file_data import FileData, SourceIdentifiers


@pytest.fixture
def onedrive_test_folder() -> str:
    """
    Pytest fixture that creates a test folder in OneDrive and deletes it after test run.
    """
    connection_config = get_connection_config()
    user_pname = connection_config.user_pname

    # Get the OneDrive client
    client: GraphClient = connection_config.get_client()
    drive = client.users[user_pname].drive

    # Generate a unique test folder path
    test_folder_path = f"utic-test-output-{uuid.uuid4()}"

    # Create the test folder
    root = drive.root
    folder = root.create_folder(test_folder_path).execute_query()
    print(f"created folder: {folder.name}")
    try:
        yield test_folder_path
    finally:
        # Teardown: delete the test folder and its contents
        folder.delete_object().execute_query()
        print(f"successfully deleted folder: {folder.name}")


def get_connection_config():
    """
    Pytest fixture that provides the OnedriveConnectionConfig for tests.
    """
    client_id = os.getenv("MS_CLIENT_ID")
    client_secret = os.getenv("MS_CLIENT_CRED")
    tenant_id = os.getenv("MS_TENANT_ID")
    user_pname = os.getenv("MS_USER_PNAME")

    connection_config = OnedriveConnectionConfig(
        client_id=client_id,
        tenant=tenant_id,
        user_pname=user_pname,
        access_config=OnedriveAccessConfig(client_cred=client_secret),
    )
    return connection_config


@pytest.mark.asyncio
@pytest.mark.tags(CONNECTOR_TYPE, SOURCE_TAG, BLOB_STORAGE_TAG)
@requires_env("MS_CLIENT_CRED", "MS_CLIENT_ID", "MS_TENANT_ID", "MS_USER_PNAME")
async def test_onedrive_source(temp_dir):
    connection_config = get_connection_config()
    index_config = OnedriveIndexerConfig(recursive=True, path="eml")

    download_config = OnedriveDownloaderConfig(download_dir=temp_dir)

    # Instantiate indexer and downloader
    indexer = OnedriveIndexer(
        connection_config=connection_config,
        index_config=index_config,
    )
    downloader = OnedriveDownloader(
        connection_config=connection_config,
        download_config=download_config,
    )

    # Run the source connector validation
    await source_connector_validation(
        indexer=indexer,
        downloader=downloader,
        configs=SourceValidationConfigs(
            test_id="onedrive",
            expected_num_files=1,
            validate_downloaded_files=True,
            exclude_fields_extend=[
                "metadata.date_created",
                "metadata.date_modified",
                "additional_metadata.LastModified",
                "additional_metadata.@microsoft.graph.downloadUrl",
            ],
        ),
    )


@pytest.mark.tags(CONNECTOR_TYPE, DESTINATION_TAG, BLOB_STORAGE_TAG)
@requires_env("MS_CLIENT_CRED", "MS_CLIENT_ID", "MS_TENANT_ID", "MS_USER_PNAME")
def xtest_onedrive_destination(upload_file: Path, onedrive_test_folder: str):
    """
    Integration test for the OneDrive destination connector.

    This test uploads a file to OneDrive and verifies that it exists.
    """
    connection_config = get_connection_config()
    # Retrieve user principal name from the connection config
    user_pname = connection_config.user_pname

    # The test folder is provided by the fixture
    destination_folder = onedrive_test_folder
    destination_fullpath = f"{destination_folder}/{upload_file.name}"

    # Configure the uploader with remote_url
    upload_config = OnedriveUploaderConfig(remote_url=f"onedrive://{destination_folder}")

    uploader = OnedriveUploader(
        connection_config=connection_config,
        upload_config=upload_config,
    )

    file_data = FileData(
        source_identifiers=SourceIdentifiers(
            fullpath=destination_fullpath,
            filename=upload_file.name,
        ),
        connector_type=CONNECTOR_TYPE,
        identifier="mock_file_data",
    )
    uploader.precheck()
    uploader.run(path=upload_file, file_data=file_data)

    # Verify that the file was uploaded
    client = connection_config.get_client()
    drive = client.users[user_pname].drive

    # Workaround: File should not have .json in the metadata.filename it comes from embedder
    uploaded_file = (
        drive.root.get_by_path(f"{destination_fullpath}.json")
        .select(["id", "name"])
        .get()
        .execute_query()
    )

    # Check if the file exists
    assert uploaded_file is not None
    assert uploaded_file.name == f"{upload_file.name}.json"



from abc import ABC
from dataclasses import dataclass
from pathlib import Path
from typing import Any, TypeVar

from pydantic import BaseModel

from unstructured_ingest.utils import ndjson
from unstructured_ingest.utils.data_prep import get_data, write_data
from unstructured_ingest.v2.interfaces import BaseProcess
from unstructured_ingest.v2.types.file_data import FileData


class UploadStagerConfig(BaseModel):
    pass


UploadStagerConfigT = TypeVar("UploadStagerConfigT", bound=UploadStagerConfig)


@dataclass
class UploadStager(BaseProcess, ABC):
    upload_stager_config: UploadStagerConfigT

    def conform_dict(self, element_dict: dict, file_data: FileData) -> dict:
        return element_dict

    def get_output_path(self, output_filename: str, output_dir: Path) -> Path:
        output_path = Path(output_filename)
        output_filename = f"{Path(output_filename).stem}{output_path.suffix}"
        output_path = Path(output_dir) / Path(f"{output_filename}")
        output_path.parent.mkdir(parents=True, exist_ok=True)
        return output_path

    def stream_update(self, input_file: Path, output_file: Path, file_data: FileData) -> None:
        with input_file.open() as in_f:
            reader = ndjson.reader(in_f)
            with output_file.open("w") as out_f:
                writer = ndjson.writer(out_f)
                for element in reader:
                    conformed_element = self.conform_dict(element_dict=element, file_data=file_data)
                    writer.write(row=conformed_element)
                    writer.f.flush()

    def process_whole(self, input_file: Path, output_file: Path, file_data: FileData) -> None:
        elements_contents = get_data(path=input_file)

        conformed_elements = [
            self.conform_dict(element_dict=element, file_data=file_data)
            for element in elements_contents
        ]
        write_data(path=output_file, data=conformed_elements)

    def run(
        self,
        elements_filepath: Path,
        file_data: FileData,
        output_dir: Path,
        output_filename: str,
        **kwargs: Any,
    ) -> Path:
        output_file = self.get_output_path(output_filename=output_filename, output_dir=output_dir)
        if elements_filepath.suffix == ".ndjson":
            self.stream_update(
                input_file=elements_filepath, output_file=output_file, file_data=file_data
            )
        elif elements_filepath.suffix == ".json":
            self.process_whole(
                input_file=elements_filepath, output_file=output_file, file_data=file_data
            )
        else:
            raise ValueError(f"Unsupported file extension: {elements_filepath}")
        return output_file

    async def run_async(
        self,
        elements_filepath: Path,
        file_data: FileData,
        output_dir: Path,
        output_filename: str,
        **kwargs: Any,
    ) -> Path:
        return self.run(
            elements_filepath=elements_filepath,
            output_dir=output_dir,
            output_filename=output_filename,
            file_data=file_data,
            **kwargs,
        )



import json
import sqlite3
import tempfile
from pathlib import Path

import pytest
from _pytest.fixtures import TopRequest

from test.integration.connectors.utils.constants import (
    DESTINATION_TAG,
    SOURCE_TAG,
    SQL_TAG,
    env_setup_path,
)
from test.integration.connectors.utils.validation.destination import (
    StagerValidationConfigs,
    stager_validation,
)
from test.integration.connectors.utils.validation.source import (
    SourceValidationConfigs,
    source_connector_validation,
)
from unstructured_ingest.v2.processes.connectors.sql.sqlite import (
    CONNECTOR_TYPE,
    SQLiteConnectionConfig,
    SQLiteDownloader,
    SQLiteDownloaderConfig,
    SQLiteIndexer,
    SQLiteIndexerConfig,
    SQLiteUploader,
    SQLiteUploadStager,
)
from unstructured_ingest.v2.types.file_data import FileData, SourceIdentifiers

SEED_DATA_ROWS = 10


@pytest.fixture
def source_database_setup() -> Path:
    with tempfile.TemporaryDirectory() as tmpdir:
        db_path = Path(tmpdir) / "mock_database.db"
        db_init_path = env_setup_path / "sql" / "sqlite" / "source" / "sqlite-schema.sql"
        assert db_init_path.exists()
        assert db_init_path.is_file()
        with sqlite3.connect(database=db_path) as sqlite_connection:
            cursor = sqlite_connection.cursor()
            with db_init_path.open("r") as f:
                query = f.read()
            cursor.executescript(query)
            for i in range(SEED_DATA_ROWS):
                sql_statment = f"INSERT INTO cars (brand, price) " f"VALUES ('brand{i}', {i})"
                cursor.execute(sql_statment)

            sqlite_connection.commit()
            cursor.close()
        yield db_path


@pytest.mark.asyncio
@pytest.mark.tags(CONNECTOR_TYPE, SOURCE_TAG, SQL_TAG)
async def test_sqlite_source(source_database_setup: Path, temp_dir: Path):
    connection_config = SQLiteConnectionConfig(database_path=source_database_setup)
    indexer = SQLiteIndexer(
        connection_config=connection_config,
        index_config=SQLiteIndexerConfig(table_name="cars", id_column="car_id", batch_size=6),
    )
    downloader = SQLiteDownloader(
        connection_config=connection_config,
        download_config=SQLiteDownloaderConfig(fields=["car_id", "brand"], download_dir=temp_dir),
    )
    await source_connector_validation(
        indexer=indexer,
        downloader=downloader,
        configs=SourceValidationConfigs(
            test_id="sqlite",
            expected_num_files=SEED_DATA_ROWS,
            expected_number_indexed_file_data=2,
            validate_downloaded_files=True,
        ),
    )


@pytest.fixture
def destination_database_setup(temp_dir: Path) -> Path:
    # Provision the local file that sqlite points to to have the desired schema for the integration
    # tests and make sure the file and connection get cleaned up by using a context manager.
    db_path = temp_dir / "elements.db"
    db_init_path = env_setup_path / "sql" / "sqlite" / "destination" / "sqlite-schema.sql"
    assert db_init_path.exists()
    assert db_init_path.is_file()
    with sqlite3.connect(database=db_path) as sqlite_connection:
        with db_init_path.open("r") as f:
            query = f.read()
        cursor = sqlite_connection.cursor()
        cursor.executescript(query)
    return db_path


def validate_destination(db_path: Path, expected_num_elements: int):
    # Run the following validations:
    # * Check that the number of records in the table match the expected value
    connection = None
    try:
        connection = sqlite3.connect(database=db_path)
        query = "select count(*) from elements;"
        cursor = connection.cursor()
        cursor.execute(query)
        count = cursor.fetchone()[0]
        assert (
            count == expected_num_elements
        ), f"dest check failed: got {count}, expected {expected_num_elements}"
    finally:
        if connection:
            connection.close()


@pytest.mark.asyncio
@pytest.mark.tags(CONNECTOR_TYPE, DESTINATION_TAG, SQL_TAG)
async def test_sqlite_destination(
    upload_file: Path, temp_dir: Path, destination_database_setup: Path
):
    # the sqlite destination connector doesn't leverage the file data but is required as an input,
    # mocking it with arbitrary values to meet the base requirements:
    mock_file_data = FileData(
        identifier="mock file data",
        connector_type=CONNECTOR_TYPE,
        source_identifiers=SourceIdentifiers(filename=upload_file.name, fullpath=upload_file.name),
    )
    stager = SQLiteUploadStager()
    staged_path = stager.run(
        elements_filepath=upload_file,
        file_data=mock_file_data,
        output_dir=temp_dir,
        output_filename=upload_file.name,
    )

    # The stager should append the `.json` suffix to the output filename passed in.
    assert staged_path.suffix == upload_file.suffix

    uploader = SQLiteUploader(
        connection_config=SQLiteConnectionConfig(database_path=destination_database_setup)
    )
    uploader.precheck()
    uploader.run(path=staged_path, file_data=mock_file_data)

    with staged_path.open("r") as f:
        staged_data = json.load(f)
    validate_destination(db_path=destination_database_setup, expected_num_elements=len(staged_data))

    uploader.run(path=staged_path, file_data=mock_file_data)
    validate_destination(db_path=destination_database_setup, expected_num_elements=len(staged_data))


@pytest.mark.tags(CONNECTOR_TYPE, DESTINATION_TAG, SQL_TAG)
@pytest.mark.parametrize("upload_file_str", ["upload_file_ndjson", "upload_file"])
def test_sqlite_stager(
    request: TopRequest,
    upload_file_str: str,
    tmp_path: Path,
):
    upload_file: Path = request.getfixturevalue(upload_file_str)
    stager = SQLiteUploadStager()
    stager_validation(
        configs=StagerValidationConfigs(test_id=CONNECTOR_TYPE, expected_count=22),
        input_file=upload_file,
        stager=stager,
        tmp_dir=tmp_path,
    )



import typing as t
from dataclasses import dataclass

import click

from unstructured_ingest.cli.interfaces import (
    CliConfig,
)
from unstructured_ingest.connector.qdrant import QdrantWriteConfig, SimpleQdrantConfig


@dataclass
class QdrantCliConfig(SimpleQdrantConfig, CliConfig):
    @staticmethod
    def get_cli_options() -> t.List[click.Option]:
        options = [
            click.Option(
                ["--collection-name"],
                required=True,
                type=str,
                help="The name of the Qdrant collection to use.",
            ),
            click.Option(
                ["--location"],
                type=str,
                help="The location of the Qdrant cluster.",
            ),
            click.Option(
                ["--url"],
                type=str,
                help="The location of the Qdrant cluster.",
            ),
            click.Option(
                ["--port"],
                type=int,
                default=6333,
                help="Port of the REST API interface. Default: 6333.",
            ),
            click.Option(
                ["--grpc-port"],
                type=int,
                default=6334,
                help="Port of the gRPC interface. Default: 6334.",
            ),
            click.Option(
                ["--prefer-grpc"],
                type=bool,
                is_flag=True,
                help="Whether to use gPRC interface whenever possible in methods. Default: False.",
            ),
            click.Option(
                ["--https"],
                type=bool,
                is_flag=True,
                help="Whether to use HTTPS(SSL) protocol. Default: False.",
            ),
            click.Option(
                ["--prefix"],
                type=str,
                help="Prefix to add the REST API endpoints.",
            ),
            click.Option(
                ["--timeout"],
                type=int,
                help="Timeout for operations. Default: 5.0 seconds for REST, unlimited for gRPC.",
            ),
            click.Option(
                ["--host"],
                type=str,
                help="Host name of the Qdrant service.",
            ),
            click.Option(
                ["--path"],
                type=str,
                help="Persistence path for QdrantLocal.",
            ),
            click.Option(
                ["--force-disable-check-same-thread"],
                type=bool,
                is_flag=True,
                help="Whether to force disable check same thread for QdrantLocal.",
            ),
            click.Option(
                ["--api-key"],
                type=str,
                help="API key for authentication in Qdrant Cloud. Default: None.",
                envvar="QDRANT_API_KEY",
                show_envvar=True,
            ),
        ]
        return options


@dataclass
class QdrantCliWriteConfig(QdrantWriteConfig, CliConfig):
    @staticmethod
    def get_cli_options() -> t.List[click.Option]:
        options = [
            click.Option(
                ["--batch-size"],
                default=50,
                type=int,
                help="Number of points to upload per batch",
            ),
            click.Option(
                ["--num-processes"],
                default=2,
                type=int,
                help="Number of parallel processes with which to upload",
            ),
        ]
        return options


def get_base_dest_cmd():
    from unstructured_ingest.cli.base.dest import BaseDestCmd

    cmd_cls = BaseDestCmd(
        cmd_name="qdrant",
        cli_config=QdrantCliConfig,
        additional_cli_options=[QdrantCliWriteConfig],
        write_config=QdrantWriteConfig,
    )
    return cmd_cls



import os
from pathlib import Path

import pytest

from test.integration.connectors.utils.constants import SOURCE_TAG, UNCATEGORIZED_TAG
from test.integration.connectors.utils.validation.source import (
    SourceValidationConfigs,
    source_connector_validation,
)
from test.integration.utils import requires_env
from unstructured_ingest.v2.errors import UserAuthError
from unstructured_ingest.v2.processes.connectors.zendesk.zendesk import (
    CONNECTOR_TYPE,
    ZendeskAccessConfig,
    ZendeskConnectionConfig,
    ZendeskDownloader,
    ZendeskDownloaderConfig,
    ZendeskIndexer,
    ZendeskIndexerConfig,
)

SUBDOMAIN = "unstructuredhelp"
EMAIL = "test@unstructured.io"


@pytest.mark.asyncio
@pytest.mark.tags(SOURCE_TAG, CONNECTOR_TYPE, UNCATEGORIZED_TAG)
@requires_env("ZENDESK_TOKEN")
async def test_zendesk_source_tickets(temp_dir: Path):
    access_config = ZendeskAccessConfig(api_token=os.environ["ZENDESK_TOKEN"])
    connection_config = ZendeskConnectionConfig(
        subdomain=SUBDOMAIN, email=EMAIL, access_config=access_config
    )

    index_config = ZendeskIndexerConfig(item_type="tickets")

    indexer = ZendeskIndexer(
        connection_config=connection_config,
        index_config=index_config,
        connector_type=CONNECTOR_TYPE,
    )

    # handle downloader.
    download_config = ZendeskDownloaderConfig(download_dir=temp_dir)

    downloader = ZendeskDownloader(
        connection_config=connection_config,
        download_config=download_config,
        connector_type=CONNECTOR_TYPE,
    )

    # Run the source connector validation
    await source_connector_validation(
        indexer=indexer,
        downloader=downloader,
        configs=SourceValidationConfigs(
            test_id="zendesk-tickets",
            expected_num_files=8,
            validate_file_data=False,
            validate_downloaded_files=True,
        ),
    )


@pytest.mark.asyncio
@pytest.mark.tags(SOURCE_TAG, CONNECTOR_TYPE, UNCATEGORIZED_TAG)
@requires_env("ZENDESK_TOKEN")
async def test_zendesk_source_articles(temp_dir):
    access_config = ZendeskAccessConfig(api_token=os.environ["ZENDESK_TOKEN"])
    connection_config = ZendeskConnectionConfig(
        subdomain=SUBDOMAIN, email=EMAIL, access_config=access_config
    )

    index_config = ZendeskIndexerConfig(item_type="articles")

    indexer = ZendeskIndexer(
        connection_config=connection_config,
        index_config=index_config,
        connector_type=CONNECTOR_TYPE,
    )

    # handle downloader.
    download_config = ZendeskDownloaderConfig(download_dir=temp_dir, extract_images=True)

    downloader = ZendeskDownloader(
        connection_config=connection_config,
        download_config=download_config,
        connector_type=CONNECTOR_TYPE,
    )

    # Run the source connector validation
    await source_connector_validation(
        indexer=indexer,
        downloader=downloader,
        configs=SourceValidationConfigs(
            test_id="zendesk-articles",
            expected_num_files=8,
            validate_file_data=True,
            validate_downloaded_files=True,
        ),
    )


@pytest.mark.tags(SOURCE_TAG, CONNECTOR_TYPE, UNCATEGORIZED_TAG)
def test_zendesk_source_articles_fail(temp_dir):
    access_config = ZendeskAccessConfig(api_token="FAKE_TOKEN")
    connection_config = ZendeskConnectionConfig(
        subdomain=SUBDOMAIN, email=EMAIL, access_config=access_config
    )

    index_config = ZendeskIndexerConfig(item_type="tickets")

    indexer = ZendeskIndexer(
        connection_config=connection_config,
        index_config=index_config,
        connector_type=CONNECTOR_TYPE,
    )
    with pytest.raises(expected_exception=UserAuthError):
        indexer.precheck()



from __future__ import annotations

from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Any, Generator, Optional
from urllib.parse import urlparse

from pydantic import Field, Secret, model_validator

from unstructured_ingest.error import SourceConnectionError
from unstructured_ingest.utils.dep_check import requires_dependencies
from unstructured_ingest.v2.interfaces import (
    AccessConfig,
    ConnectionConfig,
    Downloader,
    DownloaderConfig,
    DownloadResponse,
    Indexer,
    IndexerConfig,
)
from unstructured_ingest.v2.logger import logger
from unstructured_ingest.v2.processes.connector_registry import SourceRegistryEntry
from unstructured_ingest.v2.types.file_data import (
    FileData,
    FileDataSourceMetadata,
    SourceIdentifiers,
)

CONNECTOR_TYPE = "gitlab"
if TYPE_CHECKING:
    from gitlab import Gitlab
    from gitlab.v4.objects.projects import Project


class GitLabAccessConfig(AccessConfig):
    access_token: Optional[str] = Field(
        default=None,
        description="Optional personal access token for authenticating with the GitLab API.",
    )


class GitLabConnectionConfig(ConnectionConfig):
    access_config: Secret[GitLabAccessConfig] = Field(
        default_factory=GitLabAccessConfig,
        validate_default=True,
        description="Secret configuration for accessing the GitLab API by authentication token.",
    )
    url: str = Field(description="The full URL to the GitLab project or repository.")
    base_url: str = Field(
        default="https://gitlab.com",
        description="The base URL for the GitLab instance (default is GitLab's public domain).",
    )
    repo_path: str = Field(
        default=None,
        init=False,
        repr=False,
        description="The normalized path extracted from the repository URL.",
    )

    @model_validator(mode="after")
    def set_repo_path(self):
        """
        Parses the provided GitLab URL to extract the `base_url` and `repo_path`,
        ensuring both are properly formatted for use.

        If the URL contains a scheme (e.g., 'https') and a network location (e.g., 'gitlab.com'),
        the `base_url` is set accordingly. The repository path is extracted and normalized
        by removing any leading slashes.

        Notes:
            - If the URL contains both a scheme and network location, the `base_url` is
              extracted directly from the URL.
            - The `repo_path` is adjusted to remove any leading slashes.
            - This method assumes that the URL follows GitLab's structure
              (e.g., 'https://gitlab.com/owner/repo').
        """
        parsed_gh_url = urlparse(self.url)

        if parsed_gh_url.scheme and parsed_gh_url.netloc:
            self.base_url = f"{parsed_gh_url.scheme}://{parsed_gh_url.netloc}"
        self.repo_path = parsed_gh_url.path.lstrip("/")

        return self

    @SourceConnectionError.wrap
    @requires_dependencies(["gitlab"], extras="gitlab")
    @contextmanager
    def get_client(self) -> Generator["Gitlab", None, None]:
        from gitlab import Gitlab

        logger.info(f"Connection to GitLab: {self.base_url!r}")
        with Gitlab(
            self.base_url, private_token=self.access_config.get_secret_value().access_token
        ) as client:
            yield client

    @contextmanager
    def get_project(self) -> Generator["Project", None, None]:
        """Retrieves the specified GitLab project using the configured base URL and access token.

        Returns:
            Project: A GitLab `Project` object representing the specified repository.

        Raises:
            SourceConnectionError: If the GitLab API connection fails.
            gitlab.exceptions.GitlabGetError: If the project is not found.
        """
        with self.get_client() as client:
            logger.info(f"Accessing Project: '{self.repo_path}'")
            project = client.projects.get(self.repo_path)

            logger.info(f"Successfully accessed project '{self.repo_path}'")
            yield project


class GitLabIndexerConfig(IndexerConfig):
    path: Path = Field(
        default="/", description=("Path to the location in the repository that will be processed.")
    )
    recursive: bool = Field(
        default=True,
        description=(
            "Flag to control recursive operations when indexing. "
            "If True, the indexer will traverse directories recursively."
        ),
    )
    git_branch: Optional[str] = Field(
        default=None,
        description="The name of the branch to interact with.",
    )


@dataclass
class GitLabIndexer(Indexer):
    connection_config: GitLabConnectionConfig
    index_config: GitLabIndexerConfig

    def precheck(self) -> None:
        """Validates the connection to the GitLab instance by authenticating or
        accessing the project.

        This method ensures that the GitLab credentials and configuration are correct by
        either authenticating or attempting to fetch the specified project.

        Raises:
            SourceConnectionError: If the connection or authentication with GitLab fails.
        """

        try:
            with self.connection_config.get_client() as client:
                if self.connection_config.access_config.get_secret_value().access_token is not None:
                    client.auth()
                else:
                    client.projects.get(self.connection_config.repo_path)

        except Exception as e:
            logger.error(f"Failed to validate connection: {e}", exc_info=True)
            raise SourceConnectionError(f"Failed to validate connection: {e}")

    def run(self, **kwargs: Any) -> Generator[FileData, None, None]:
        """Iterates over the GitLab repository tree and yields file metadata as `FileData` objects.

        This method fetches the repository tree for the specified branch and iterates
        over its contents. For each file (blob), it generates a `FileData` object containing
        the file's metadata, path, and permissions.

        Args:
            **kwargs (Any): Additional keyword arguments (if required).

        Yields:
            FileData: A generator that yields `FileData` objects representing each file (blob)
            in the repository.
        """
        with self.connection_config.get_project() as project:
            ref = self.index_config.git_branch or project.default_branch

            files = project.repository_tree(
                path=str(self.index_config.path),
                ref=ref,
                recursive=self.index_config.recursive,
                iterator=True,
                all=True,
            )

        for file in files:
            relative_path = str(Path(file["path"]).relative_to(self.index_config.path))
            if file["type"] == "blob":
                record_locator = {
                    "file_path": file["path"],
                    "ref": ref,
                }

                yield FileData(
                    identifier=file["id"],
                    connector_type=CONNECTOR_TYPE,
                    source_identifiers=SourceIdentifiers(
                        fullpath=file["path"],
                        filename=Path(file["path"]).name,
                        rel_path=relative_path,
                    ),
                    metadata=FileDataSourceMetadata(
                        url=file["id"],
                        record_locator=record_locator,
                        permissions_data=[{"mode": file["mode"]}],
                    ),
                    additional_metadata={},
                )


class GitLabDownloaderConfig(DownloaderConfig):
    pass


@dataclass
class GitLabDownloader(Downloader):
    connection_config: GitLabConnectionConfig
    download_config: GitLabDownloaderConfig

    def run(self, file_data: FileData, **kwargs: Any) -> DownloadResponse:
        """Downloads a file from the repository and returns a `DownloadResponse`.

        Args:
            file_data (FileData): Metadata about the file to be downloaded.
            **kwargs (Any): Additional arguments (if required).

        Returns:
            DownloadResponse: A response object containing the download details.
        """
        download_path = self.get_download_path(file_data=file_data)
        if download_path is None:
            logger.error(
                "Generated download path is None, source_identifiers might be missing"
                "from FileData."
            )
            raise ValueError("Generated invalid download path.")

        self._download_file(file_data, download_path)
        return self.generate_download_response(file_data=file_data, download_path=download_path)

    def _download_file(self, file_data: FileData, download_path: Path) -> None:
        # NOTE: Indexer should supply the record locator in metadata
        if (
            file_data.metadata.record_locator is None
            or "ref" not in file_data.metadata.record_locator
            or "file_path" not in file_data.metadata.record_locator
        ):
            logger.error(
                f"Invalid record locator in metadata: {file_data.metadata.record_locator}."
                "Keys 'ref' and 'path' must be present."
            )
            raise ValueError("Invalid record locator.")

        ref = file_data.metadata.record_locator["ref"]
        path = file_data.metadata.record_locator["file_path"]
        download_path.parent.mkdir(exist_ok=True, parents=True)

        with self.connection_config.get_project() as project:
            project_file = project.files.get(file_path=path, ref=ref)
            with open(download_path, "wb") as file:
                file.write(project_file.decode())


gitlab_source_entry = SourceRegistryEntry(
    connection_config=GitLabConnectionConfig,
    indexer_config=GitLabIndexerConfig,
    indexer=GitLabIndexer,
    downloader_config=GitLabDownloaderConfig,
    downloader=GitLabDownloader,
)
